import pytest
from ase.build import molecule
from gpaw import GPAW, restart
from gpaw.test import equal


@pytest.mark.mom
def test_mom_lcao_smearing(in_tmp_dir):
    atoms = molecule('CO')
    atoms.center(vacuum=2)

    calc = GPAW(mode='lcao',
                basis='dzp',
                nbands=7,
                h=0.24,
                xc='PBE',
                spinpol=True,
                symmetry='off',
                convergence={'energy': 100,
                             'density': 1e-3})

    atoms.calc = calc
    # Ground-state calculation
    E_gs = atoms.get_potential_energy()

    calc.write('co_lcao_gs.gpw', 'all')

    f_sn = []
    for spin in range(calc.get_number_of_spins()):
        f_n = calc.get_occupation_numbers(spin=spin)
        f_sn.append(f_n)

    ne0_gs = f_sn[0].sum()
    f_sn[0][3] -= 1.
    f_sn[0][5] += 1.

    # Test both MOM and fixed occupations with Gaussian smearing
    for i in [True, False]:
        atoms, calc = restart('co_lcao_gs.gpw', txt='-')

        # Excited-state calculation with Gaussian
        # smearing of the occupation numbers
        calc.set(occupations={'name': 'mom', 'numbers': f_sn,
                              'width': 0.01, 'use_fixed_occupations': i})
        E_es = atoms.get_potential_energy()

        f_n0 = calc.get_occupation_numbers(spin=0)
        ne0_es = f_n0.sum()

        dE = E_es - E_gs
        dne0 = ne0_es - ne0_gs
        print(dE)
        equal(dE, 9.8445603513, 0.01)
        equal(dne0, 0.0, 1e-16)



# Copyright (C) 2010  CAMd
# Copyright (C) 2010  Argonne National Laboratory
# Please see the accompanying LICENSE file for further information.
import numpy as np
from scipy.linalg import eigh

from gpaw import debug
from gpaw.matrix_descriptor import MatrixDescriptor
from gpaw.mpi import broadcast_exception
from gpaw.blacs import BlacsGrid, Redistributor
from gpaw.utilities import uncamelcase
from gpaw.utilities.blas import mmm, r2k
from gpaw.utilities.scalapack import (pblas_simple_gemm, pblas_tran,
                                      scalapack_tri2full)
from gpaw.utilities.tools import tri2full
from gpaw.utilities.timing import nulltimer


def get_KohnSham_layouts(sl, mode, gd, bd, block_comm, dtype,
                         elpasolver=None, **kwargs):
    """Create Kohn-Sham layouts object."""
    # Not needed for AtomPAW special mode, as usual we just provide whatever
    # happens to make the code not crash
    if not isinstance(mode, str):
        return None  # XXX
    name = 'OrbitalLayouts'
    args = (gd, bd, block_comm, dtype)
    if sl is not None:
        name = 'Blacs' + name
        assert len(sl) == 3
        args += tuple(sl)
        if elpasolver is not None:
            kwargs['elpasolver'] = elpasolver
    ksl = {'BlacsOrbitalLayouts': BlacsOrbitalLayouts,
           'OrbitalLayouts': OrbitalLayouts}[name](*args,
                                                   **kwargs)
    return ksl


class KohnShamLayouts:
    using_blacs = False  # This is only used by a regression test
    matrix_descriptor_class = None
    accepts_decomposed_overlap_matrix = False

    def __init__(self, gd, bd, block_comm, dtype, timer=nulltimer):
        assert gd.comm.parent is bd.comm.parent  # must have same parent comm
        self.world = bd.comm.parent
        self.gd = gd
        self.bd = bd
        self.dtype = dtype
        self.block_comm = block_comm
        self.timer = timer
        self._kwargs = {'timer': timer}

        if gd.comm.rank == 0:
            self.column_comm = bd.comm
        else:
            self.column_comm = None

    def get_keywords(self):
        return self._kwargs.copy()  # just a shallow copy...

    def diagonalize(self, *args, **kwargs):
        raise RuntimeError('Virtual member function should not be called.')

    def inverse_cholesky(self, *args, **kwargs):
        raise RuntimeError('Virtual member function should not be called.')

    def new_descriptor(self):
        return self.matrix_descriptor_class(self.bd, self.gd, self)

    def __repr__(self):
        return uncamelcase(self.__class__.__name__)

    def get_description(self):
        """Description of this object in prose, e.g. for logging.

        Subclasses are expected to override this with something useful."""
        return repr(self)


class BlacsLayouts(KohnShamLayouts):
    using_blacs = True  # This is only used by a regression test

    def __init__(self, gd, bd, block_comm, dtype, mcpus, ncpus,
                 blocksize, timer=nulltimer):
        KohnShamLayouts.__init__(self, gd, bd, block_comm, dtype,
                                 timer)
        # WARNING: Do not create the BlacsGrid on a communicator which does not
        # contain block_comm.rank = 0. This will break BlacsBandLayouts which
        # assume eps_M will be broadcast over block_comm.
        self.blocksize = blocksize
        self.blockgrid = BlacsGrid(self.block_comm, mcpus, ncpus)

    def get_description(self):
        title = 'BLACS'
        template = '%d x %d grid with %d x %d blocksize'
        return (title, template)


class BlacsOrbitalLayouts(BlacsLayouts):
    """ScaLAPACK Dense Linear Algebra.

    This class is instantiated in LCAO.  Not for casual use, at least for now.

    Requires two distributors and three descriptors for initialization
    as well as grid descriptors and band descriptors. Distributors are
    for cols2blocks (1D -> 2D BLACS grid) and blocks2cols (2D -> 1D
    BLACS grid). ScaLAPACK operations must occur on 2D BLACS grid for
    performance and scalability.

    _general_diagonalize is "hard-coded" for LCAO.
    Expects both Hamiltonian and Overlap matrix to be on the 2D BLACS grid.
    This is done early on to save memory.
    """
    # XXX rewrite this docstring a bit!

    # This class 'describes' all the LCAO Blacs-related layouts
    def __init__(self, gd, bd, block_comm, dtype, mcpus, ncpus,
                 blocksize, nao, elpasolver=None, timer=nulltimer):
        BlacsLayouts.__init__(self, gd, bd, block_comm, dtype,
                              mcpus, ncpus, blocksize, timer)
        nbands = bd.nbands
        self.blocksize = blocksize

        self.orbital_comm = self.bd.comm
        self.naoblocksize = naoblocksize = -((-nao) // self.orbital_comm.size)
        self.nao = nao

        # Range of basis functions for BLACS distribution of matrices:
        self.Mmax = nao
        self.Mstart = min(bd.comm.rank * naoblocksize, self.Mmax)
        self.Mstop = min(self.Mstart + naoblocksize, self.Mmax)
        self.mynao = self.Mstop - self.Mstart

        # Column layout for one matrix per band rank:
        self.columngrid = BlacsGrid(bd.comm, bd.comm.size, 1)
        self.mMdescriptor = self.columngrid.new_descriptor(nao, nao,
                                                           naoblocksize, nao)
        self.nMdescriptor = self.columngrid.new_descriptor(nbands, nao,
                                                           bd.maxmynbands, nao)

        # parallelprint(world, (mynao, self.mMdescriptor.shape))

        # Column layout for one matrix in total (only on grid masters):
        self.single_column_grid = BlacsGrid(self.column_comm, bd.comm.size, 1)
        self.mM_unique_descriptor = self.single_column_grid.new_descriptor(
            nao, nao, naoblocksize, nao)

        # nM_unique_descriptor is meant to hold the coefficients after
        # diagonalization.  BLACS requires it to be nao-by-nao, but
        # we only fill meaningful data into the first nbands columns.
        #
        # The array will then be trimmed and broadcast across
        # the grid descriptor's communicator.
        self.nM_unique_descriptor = self.single_column_grid.new_descriptor(
            nbands, nao, bd.maxmynbands, nao)

        # Fully blocked grid for diagonalization with many CPUs:
        self.mmdescriptor = self.blockgrid.new_descriptor(nao, nao, blocksize,
                                                          blocksize)

        # self.nMdescriptor = nMdescriptor
        self.mM2mm = Redistributor(self.block_comm, self.mM_unique_descriptor,
                                   self.mmdescriptor)
        self.mm2nM = Redistributor(self.block_comm, self.mmdescriptor,
                                   self.nM_unique_descriptor)

        self.libelpa = None
        if elpasolver is not None:
            # XXX forward solver to libelpa
            from gpaw.utilities.elpa import LibElpa
            self.libelpa = LibElpa(self.mmdescriptor, solver=elpasolver,
                                   nev=nbands)

    @property
    def accepts_decomposed_overlap_matrix(self):
        return self.libelpa is not None

    def diagonalize(self, H_mm, C_nM, eps_n, S_mm,
                    is_already_decomposed=False):
        # C_nM needs to be simultaneously compatible with:
        # 1. outdescriptor
        # 2. broadcast with gd.comm
        # We will does this with a dummy buffer C2_nM
        outdescriptor = self.mm2nM.dstdescriptor  # blocks2cols
        blockdescriptor = self.mM2mm.dstdescriptor  # cols2blocks

        dtype = S_mm.dtype
        eps_M = np.empty(C_nM.shape[-1])  # empty helps us debug
        subM, subN = outdescriptor.gshape
        C_mm = blockdescriptor.zeros(dtype=dtype)

        self.timer.start('General diagonalize')
        # general_diagonalize_ex may have a buffer overflow, so
        # we no longer use it
        # blockdescriptor.general_diagonalize_ex(H_mm, S_mm.copy(), C_mm,
        #                                        eps_M,
        #                                        UL='L', iu=self.bd.nbands)
        if self.libelpa is not None:
            assert blockdescriptor is self.libelpa.desc
            scalapack_tri2full(blockdescriptor, H_mm)

            # elpa will write decomposed form of S_mm into S_mm.
            # Other KSL diagonalization functions do *not* overwrite S_mm.
            self.libelpa.general_diagonalize(
                H_mm, S_mm, C_mm, eps_M[:self.bd.nbands],
                is_already_decomposed=is_already_decomposed)
        else:
            blockdescriptor.general_diagonalize_dc(H_mm, S_mm.copy(), C_mm,
                                                   eps_M, UL='L')
        self.timer.stop('General diagonalize')

        # Make C_nM compatible with the redistributor
        self.timer.start('Redistribute coefs')
        if outdescriptor:
            C2_nM = C_nM
        else:
            C2_nM = outdescriptor.empty(dtype=dtype)
        assert outdescriptor.check(C2_nM)
        self.mm2nM.redistribute(C_mm, C2_nM, subM, subN)  # blocks2cols
        self.timer.stop('Redistribute coefs')

        self.timer.start('Send coefs to domains')
        # eps_M is already on block_comm.rank = 0
        # easier to broadcast eps_M to all and
        # get the correct slice afterward.
        self.block_comm.broadcast(eps_M, 0)
        eps_n[:] = eps_M[self.bd.get_slice()]
        self.gd.comm.broadcast(C_nM, 0)
        self.timer.stop('Send coefs to domains')

    def distribute_overlap_matrix(self, S_qmM, root=0,
                                  add_hermitian_conjugate=False):
        # Some MPI implementations need a lot of memory to do large
        # reductions.  To avoid trouble, we do comm.sum on smaller blocks
        # of S (this code is also safe for arrays smaller than blocksize)
        Sflat_x = S_qmM.ravel()
        blocksize = 2**23 // Sflat_x.itemsize  # 8 MiB
        nblocks = -(-len(Sflat_x) // blocksize)
        Mstart = 0
        self.timer.start('blocked summation')
        for i in range(nblocks):
            self.gd.comm.sum(Sflat_x[Mstart:Mstart + blocksize], root=root)
            Mstart += blocksize
        assert Mstart + blocksize >= len(Sflat_x)
        self.timer.stop('blocked summation')

        xshape = S_qmM.shape[:-2]
        if len(xshape) == 0:
            S_qmM = S_qmM[np.newaxis]
        assert S_qmM.ndim == 3

        blockdesc = self.mmdescriptor
        coldesc = self.mM_unique_descriptor
        S_qmm = blockdesc.zeros(len(S_qmM), S_qmM.dtype)

        if not coldesc:  # XXX ugly way to sort out inactive ranks
            S_qmM = coldesc.zeros(len(S_qmM), S_qmM.dtype)

        self.timer.start('Scalapack redistribute')
        for S_mM, S_mm in zip(S_qmM, S_qmm):
            self.mM2mm.redistribute(S_mM, S_mm)
            if add_hermitian_conjugate:
                if blockdesc.active:
                    pblas_tran(1.0, S_mm.copy(), 1.0, S_mm,
                               blockdesc, blockdesc)

            if self.libelpa is not None:
                # Elpa needs the full H_mm, but apparently does not
                # need the full S_mm.  However that fact is not documented,
                # and hence we stay on the safe side, tri2full-ing the
                # matrix.
                scalapack_tri2full(blockdesc, S_mm)

        self.timer.stop('Scalapack redistribute')
        return S_qmm.reshape(xshape + blockdesc.shape)

    def get_overlap_matrix_shape(self):
        return self.mmdescriptor.shape

    def calculate_blocked_density_matrix(self, f_n, C_nM):
        nbands = self.bd.nbands
        nao = self.nao
        dtype = C_nM.dtype

        self.nMdescriptor.checkassert(C_nM)
        if self.gd.rank == 0:
            Cf_nM = (C_nM * f_n[:, None])
        else:
            C_nM = self.nM_unique_descriptor.zeros(dtype=dtype)
            Cf_nM = self.nM_unique_descriptor.zeros(dtype=dtype)

        r = Redistributor(self.block_comm, self.nM_unique_descriptor,
                          self.mmdescriptor)

        Cf_mm = self.mmdescriptor.zeros(dtype=dtype)
        r.redistribute(Cf_nM, Cf_mm, nbands, nao)
        del Cf_nM

        C_mm = self.mmdescriptor.zeros(dtype=dtype)
        r.redistribute(C_nM, C_mm, nbands, nao)
        # no use to delete C_nM as it's in the input...

        rho_mm = self.mmdescriptor.zeros(dtype=dtype)

        if 1:  # if self.libelpa is None:
            pblas_simple_gemm(self.mmdescriptor,
                              self.mmdescriptor,
                              self.mmdescriptor,
                              Cf_mm, C_mm, rho_mm, transa='C')
        else:
            # elpa_hermitian_multiply was not faster than the ordinary
            # multiplication in the test.  The way we have things distributed,
            # we need to transpose things at the moment.
            #
            # Rather than enabling this, we should store the coefficients
            # in an appropriate 2D block cyclic format (c_nm) and not the
            # current C_nM format.  This makes it possible to avoid
            # redistributing the coefficients at all.  But we don't have time
            # to implement this at the moment.
            mul = self.libelpa.hermitian_multiply
            desc = self.mmdescriptor
            from gpaw.utilities.scalapack import pblas_tran

            def T(array):
                tmp = array.copy()
                pblas_tran(alpha=1.0, a_MN=tmp,
                           beta=0.0, c_NM=array,
                           desca=desc, descc=desc)
            T(C_mm)
            T(Cf_mm)
            mul(C_mm, Cf_mm, rho_mm,
                desc, desc, desc,
                uplo_a='X', uplo_c='X')

        return rho_mm

    def calculate_density_matrix(self, f_n, C_nM, rho_mM=None):
        """Calculate density matrix from occupations and coefficients.

        Presently this function performs the usual scalapack 3-step trick:
        redistribute-numbercrunching-backdistribute.


        Notes on future performance improvement.

        As per the current framework, C_nM exists as copies on each
        domain, i.e. this is not parallel over domains.  We'd like to
        correct this and have an efficient distribution using e.g. the
        block communicator.

        The diagonalization routine and other parts of the code should
        however be changed to accommodate the following scheme:

        Keep coefficients in C_mm form after the diagonalization.
        rho_mm can then be directly calculated from C_mm without
        redistribution, after which we only need to redistribute
        rho_mm across domains.

        """
        dtype = C_nM.dtype
        rho_mm = self.calculate_blocked_density_matrix(f_n, C_nM)
        rback = Redistributor(self.block_comm, self.mmdescriptor,
                              self.mM_unique_descriptor)
        rho1_mM = self.mM_unique_descriptor.zeros(dtype=dtype)
        rback.redistribute(rho_mm, rho1_mM)
        del rho_mm

        if rho_mM is None:
            if self.gd.rank == 0:
                rho_mM = rho1_mM
            else:
                rho_mM = self.mMdescriptor.zeros(dtype=dtype)

        self.gd.comm.broadcast(rho_mM, 0)
        return rho_mM

    def distribute_to_columns(self, rho_mm, srcdescriptor):
        redistributor = Redistributor(self.block_comm,  # XXX
                                      srcdescriptor,
                                      self.mM_unique_descriptor)
        rho_mM = redistributor.redistribute(rho_mm)
        if self.gd.rank != 0:
            rho_mM = self.mMdescriptor.zeros(dtype=rho_mm.dtype)
        self.gd.comm.broadcast(rho_mM, 0)
        return rho_mM

    def oldcalculate_density_matrix(self, f_n, C_nM, rho_mM=None):
        # This version is parallel over the band descriptor only.
        # This is inefficient, but let's keep it for a while in case
        # there's trouble with the more efficient version
        if rho_mM is None:
            rho_mM = self.mMdescriptor.zeros(dtype=C_nM.dtype)

        Cf_nM = (C_nM * f_n[:, None]).conj()
        pblas_simple_gemm(self.nMdescriptor, self.nMdescriptor,
                          self.mMdescriptor, Cf_nM, C_nM, rho_mM, transa='T')
        return rho_mM

    def get_transposed_density_matrix(self, f_n, C_nM, rho_mM=None):
        return self.calculate_density_matrix(f_n, C_nM, rho_mM).conj()

    def get_description(self):
        (title, template) = BlacsLayouts.get_description(self)
        bg = self.blockgrid
        desc = self.mmdescriptor
        s = template % (bg.nprow, bg.npcol, desc.mb, desc.nb)
        if self.libelpa is not None:
            solver = self.libelpa.description
        else:
            solver = 'ScaLAPACK'
        return ''.join([title, ' / ', solver, ', ', s])


class OrbitalLayouts(KohnShamLayouts):
    def __init__(self, gd, bd, block_comm, dtype, nao,
                 timer=nulltimer):
        KohnShamLayouts.__init__(self, gd, bd, block_comm, dtype,
                                 timer)
        self.mMdescriptor = MatrixDescriptor(nao, nao)
        self.nMdescriptor = MatrixDescriptor(bd.mynbands, nao)

        self.Mstart = 0
        self.Mstop = nao
        self.Mmax = nao
        self.mynao = nao
        self.nao = nao
        self.orbital_comm = bd.comm

    def diagonalize(self, H_MM, C_nM, eps_n, S_MM,
                    overwrite_S=False):
        assert not overwrite_S
        eps_M = np.empty(C_nM.shape[-1])
        self.block_comm.broadcast(H_MM, 0)
        self.block_comm.broadcast(S_MM, 0)
        # The result on different processor is not necessarily bit-wise
        # identical, so only domain master performs computation
        with broadcast_exception(self.gd.comm):
            if self.gd.comm.rank == 0:
                self._diagonalize(H_MM, S_MM.copy(), eps_M)
        self.gd.comm.broadcast(H_MM, 0)
        self.gd.comm.broadcast(eps_M, 0)
        eps_n[:] = eps_M[self.bd.get_slice()]
        C_nM[:] = H_MM[self.bd.get_slice()]

    def _diagonalize(self, H_MM, S_MM, eps_M):
        """Serial diagonalize via LAPACK."""
        # This is replicated computation but ultimately avoids
        # additional communication
        if eps_M.size == 0:
            return
        eps_M[:], H_MM.T[:] = eigh(H_MM, S_MM,
                                   overwrite_a=True,
                                   check_finite=debug)
        if H_MM.dtype == complex:
            np.negative(H_MM.imag, H_MM.imag)

    def estimate_memory(self, mem, dtype):
        itemsize = mem.itemsize[dtype]
        mem.subnode('eps [M]', self.nao * mem.floatsize)
        mem.subnode('H [MM]', self.nao * self.nao * itemsize)

    def distribute_overlap_matrix(self, S_qMM, root=0,
                                  add_hermitian_conjugate=False):
        self.gd.comm.sum(S_qMM, root)
        if add_hermitian_conjugate:
            S_qMM += S_qMM.swapaxes(-1, -2).conj()
        return S_qMM

    def get_overlap_matrix_shape(self):
        return self.nao, self.nao

    def calculate_density_matrix(self, f_n, C_nM, rho_MM=None, C2_nM=None):
        # Only a madman would use a non-transposed density matrix.
        # Maybe we should use the get_transposed_density_matrix instead
        if rho_MM is None:
            rho_MM = np.zeros((self.mynao, self.nao), dtype=C_nM.dtype)
        # XXX Should not conjugate, but call gemm(..., 'c')
        # Although that requires knowing C_Mn and not C_nM.
        # that also conforms better to the usual conventions in literature

        # Use only occupied bands to construct density matrix
        occupied = abs(f_n) > 1.0e-10
        if C2_nM is None:
            C2_nM = C_nM
        Cf_Mn = \
            np.ascontiguousarray(C2_nM[occupied].T.conj() * f_n[occupied])
        # gemm(1.0, C_nM[occupied], Cf_Mn, 0.0, rho_MM, 'n')
        mmm(1.0, Cf_Mn, 'N', C_nM[occupied], 'N', 0.0, rho_MM)
        return rho_MM

    def get_transposed_density_matrix(self, f_n, C_nM, rho_MM=None):
        return self.calculate_density_matrix(f_n, C_nM, rho_MM).T.copy()

        # if rho_MM is None:
        #     rho_MM = np.zeros((self.mynao, self.nao), dtype=C_nM.dtype)
        # C_Mn = C_nM.T.copy()
        # gemm(1.0, C_Mn, f_n[np.newaxis, :] * C_Mn, 0.0, rho_MM, 'c')
        # self.bd.comm.sum(rho_MM)
        # return rho_MM

    def alternative_calculate_density_matrix(self, f_n, C_nM, rho_MM=None):
        if rho_MM is None:
            rho_MM = np.zeros((self.mynao, self.nao), dtype=C_nM.dtype)
        # Alternative suggestion. Might be faster. Someone should test this
        C_Mn = C_nM.T.copy()
        r2k(0.5, C_Mn, f_n * C_Mn, 0.0, rho_MM)
        tri2full(rho_MM)
        return rho_MM

    def get_description(self):
        return 'Serial LAPACK'

    def calculate_density_matrix_delta(self, d_nn, C_nM, rho_MM=None):
        # Only a madman would use a non-transposed density matrix.
        # Maybe we should use the get_transposed_density_matrix instead
        if rho_MM is None:
            rho_MM = np.zeros((self.mynao, self.nao), dtype=C_nM.dtype)
        Cd_Mn = np.zeros((self.nao, self.bd.mynbands), dtype=C_nM.dtype)
        # XXX Should not conjugate, but call gemm(..., 'c')
        # Although that requires knowing C_Mn and not C_nM.
        # that also conforms better to the usual conventions in literature
        C_Mn = C_nM.T.conj().copy()
        # gemm(1.0, d_nn, C_Mn, 0.0, Cd_Mn, 'n')
        # gemm(1.0, C_nM, Cd_Mn, 0.0, rho_MM, 'n')
        mmm(1.0, C_Mn, 'N', d_nn, 'N', 0.0, Cd_Mn)
        mmm(1.0, Cd_Mn, 'N', C_nM, 'N', 0.0, rho_MM)
        self.bd.comm.sum(rho_MM)
        return rho_MM

    def get_transposed_density_matrix_delta(self, d_nn, C_nM, rho_MM=None):
        return self.calculate_density_matrix_delta(d_nn, C_nM, rho_MM).T.copy()



import pytest
from ase import Atoms, Atom
from ase.vibrations.placzek import Placzek
from ase.parallel import world, DummyMPI

from gpaw import GPAW
from gpaw.lrtddft.kssingle import KSSingles
from gpaw.test import equal


@pytest.mark.skip(reason='TODO')
def test_rraman():
    """
    Test Placzek Raman implementation
    """

    txt = '-'
    txt = None
    xc = 'LDA'

    # ---------------------------------------------------
    # run

    R = 0.7  # approx. experimental bond length
    a = 4.0
    c = 5.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    calc = GPAW(xc=xc, nbands=3, spinpol=False, eigensolver='rmm-diis',
                txt=txt)
    H2.calc = calc
    H2.get_potential_energy()

    gsname = exname = 'rraman'
    exkwargs = {'eps': 0.0, 'jend': 1}
    pz = Placzek(H2, KSSingles, gsname=gsname, exname=exname,
                 exkwargs=exkwargs)
    pz.run()

    # ---------------------------------------------------
    # check

    # check size
    kss = KSSingles('rraman-d0.010.eq.ex.gz')
    assert len(kss) == 1

    om = 5
    pz = Placzek(H2, KSSingles, gsname=gsname, exname=exname,
                 verbose=True,)
    ai = pz.absolute_intensity(omega=om)[-1]
    equal(ai, 301.659994669, 1e-3)  # earlier obtained value
    i = pz.intensity(omega=om)[-1]
    equal(i, 7.87430831452e-05, 1e-11)  # earlier obtained value
    pz.summary(omega=5, method='frederiksen')

    # parallel ------------------------

    if world.size > 1 and world.rank == 0:
        # single core
        comm = DummyMPI()
        pzsi = Placzek(H2, KSSingles, gsname=gsname, exname=exname,
                       comm=comm, verbose=True,)
        isi = pzsi.intensity(omega=om)[-1]
        equal(isi, i, 1e-11)



"""Test LCAO density calculation and conversion to grid.
Test that the density generated by the following three procedures is the same:

 * basis_functions.construct_density as used in a normal calculation
 * axpy used on the psit_nG as constructed by lcao_to_grid
 * axpy used on the Phit_MG[i] * Phit_MG[j] * rho[j, i], where Phit_MG
   are the actual basis functions on the grid, constructed using lcao_to_grid

TODO: non-gamma-point test

"""
import numpy as np
import pytest
from ase.build import molecule

from gpaw import GPAW, ConvergenceError
from gpaw.new.scf import SCFConvergenceError
from gpaw.utilities.blas import axpy


@pytest.mark.later
def test_lcao_density():
    system = molecule('H2O')
    system.center(vacuum=2.5)

    calc = GPAW(mode='lcao',
                maxiter=1)

    system.calc = calc
    try:
        system.get_potential_energy()
    except (ConvergenceError, SCFConvergenceError):
        pass

    wfs = calc.wfs
    kpt = wfs.kpt_u[0]
    nt_G = calc.density.gd.zeros()
    bfs = wfs.basis_functions
    nao = wfs.setups.nao
    f_n = kpt.f_n
    rho_MM = np.zeros((nao, nao))
    wfs.calculate_density_matrix(kpt.f_n, kpt.C_nM, rho_MM)

    bfs.construct_density(rho_MM, nt_G, -1)

    nbands = wfs.bd.nbands
    psit_nG = wfs.gd.zeros(nbands)
    bfs.lcao_to_grid(kpt.C_nM, psit_nG, -1)

    nt2_G = calc.density.gd.zeros()
    for f, psit_G in zip(f_n, psit_nG):
        axpy(f, psit_G**2, nt2_G)

    identity_MM = np.identity(nao)
    Phit_MG = calc.wfs.gd.zeros(nao)
    bfs.lcao_to_grid(identity_MM, Phit_MG, -1)

    nt3_G = calc.density.gd.zeros()
    for M1, Phit1_G in enumerate(Phit_MG):
        for M2, Phit2_G in enumerate(Phit_MG):
            nt3_G += rho_MM[M1, M2] * Phit1_G * Phit2_G

    err1_G = nt2_G - nt_G
    err2_G = nt3_G - nt_G

    maxerr1 = np.abs(err1_G).max()
    maxerr2 = np.abs(err2_G).max()

    print('err1', maxerr1)
    print('err2', maxerr2)

    assert max(maxerr1, maxerr2) < 1e-15



from __future__ import annotations

from functools import partial
from typing import Callable

import numpy as np
from ase.units import Ha
from gpaw.core.arrays import DistributedArrays as DA
from gpaw.core.atom_centered_functions import AtomArrays as AA
from gpaw.core.matrix import Matrix
from gpaw.gpu import as_xp
from gpaw.mpi import broadcast_float
from gpaw.new import zip
from gpaw.new.calculation import DFTState
from gpaw.new.eigensolver import Eigensolver
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions
from gpaw.typing import Array1D, Array2D
from gpaw.utilities.blas import axpy
from gpaw.yml import obj2yaml as o2y

AAFunc = Callable[[AA, AA], AA]


class Davidson(Eigensolver):
    def __init__(self,
                 nbands: int,
                 wf_grid,
                 band_comm,
                 preconditioner_factory,
                 niter=2,
                 blocksize=10,
                 converge_bands='occupied',
                 scalapack_parameters=None):
        self.niter = niter
        self.converge_bands = converge_bands

        self.H_NN = None
        self.S_NN = None
        self.M_nn = None
        self.work_arrays: np.ndarray | None = None

        self.preconditioner = preconditioner_factory(blocksize)

    def __str__(self):
        return o2y(dict(name='Davidson',
                        niter=self.niter,
                        converge_bands=self.converge_bands))

    def _initialize(self, ibzwfs):
        # First time: allocate work-arrays
        wfs = ibzwfs.wfs_qs[0][0]
        assert isinstance(wfs, PWFDWaveFunctions)
        xp = wfs.psit_nX.xp
        B = ibzwfs.nbands
        domain_comm = wfs.psit_nX.desc.comm
        band_comm = wfs.band_comm
        shape = ibzwfs.get_max_shape()
        shape = (2, B) + shape
        dtype = wfs.psit_nX.data.dtype
        self.work_arrays = xp.empty(shape, dtype)

        dtype = wfs.psit_nX.desc.dtype
        if domain_comm.rank == 0 and band_comm.rank == 0:
            self.H_NN = Matrix(2 * B, 2 * B, dtype, xp=xp)
            self.S_NN = Matrix(2 * B, 2 * B, dtype, xp=xp)
        else:
            self.H_NN = self.S_NN = Matrix(0, 0)

        self.M_nn = Matrix(B, B, dtype,
                           dist=(band_comm, band_comm.size),
                           xp=xp)

    def iterate(self, state: DFTState, hamiltonian: Hamiltonian) -> float:
        """Iterate on state given fixed hamiltonian.

        Returns
        -------
        float:
            Weighted error of residuals:::

                   ~     ~ ~
              R = (H - ε S)ψ
               n        n   n
        """

        if self.work_arrays is None:
            self._initialize(state.ibzwfs)

        assert self.M_nn is not None

        dS = state.ibzwfs.wfs_qs[0][0].setups.overlap_correction
        dH = state.potential.dH
        Ht = partial(hamiltonian.apply, state.potential.vt_sR)
        ibzwfs = state.ibzwfs
        error = 0.0

        weight_un = calculate_weights(self.converge_bands, ibzwfs)

        for wfs, weight_n in zip(ibzwfs, weight_un):
            e = self.iterate1(wfs, Ht, dH, dS, weight_n)
            error += wfs.weight * e
        return ibzwfs.kpt_comm.sum(float(error)) * ibzwfs.spin_degeneracy

    def iterate1(self, wfs, Ht, dH, dS, weight_n):
        H_NN = self.H_NN
        S_NN = self.S_NN
        M_nn = self.M_nn

        xp = M_nn.xp

        psit_nX = wfs.psit_nX
        psit2_nX = psit_nX.new(data=self.work_arrays[0])
        psit3_nX = psit_nX.new(data=self.work_arrays[1])

        B = psit_nX.dims[0]  # number of bands
        eig_N = xp.empty(2 * B)

        wfs.subspace_diagonalize(Ht, dH,
                                 work_array=psit2_nX.data,
                                 Htpsit_nX=psit3_nX)
        residual_nX = psit3_nX  # will become (H-e*S)|psit> later

        P_ani = wfs.P_ani
        P2_ani = P_ani.new()
        P3_ani = P_ani.new()

        domain_comm = psit_nX.desc.comm
        band_comm = psit_nX.comm
        is_domain_band_master = domain_comm.rank == 0 and band_comm.rank == 0

        M0_nn = M_nn
        assert band_comm.size == 1

        if domain_comm.rank == 0:
            eig_N[:B] = xp.asarray(wfs.eig_n)

        def me(a, b, function=None):
            """Matrix elements"""
            return a.matrix_elements(b,
                                     domain_sum=False,
                                     out=M_nn,
                                     function=function,
                                     cc=True)

        Ht = partial(Ht, out=residual_nX, spin=wfs.spin)
        dH = partial(dH, spin=wfs.spin)

        calculate_residuals(residual_nX, dH, dS, wfs, P2_ani, P3_ani)

        def copy(C_nn: Array2D) -> None:
            domain_comm.sum(M_nn.data, 0)
            if domain_comm.rank == 0:
                M_nn.redist(M0_nn)
                if band_comm.rank == 0:
                    C_nn[:] = M0_nn.data

        for i in range(self.niter):
            if i == self.niter - 1:  # last iteration
                # Calulate error before we destroy residuals:
                if weight_n is None:
                    error = np.inf
                else:
                    error = weight_n @ as_xp(residual_nX.norm2(), np)
                    if wfs.ncomponents == 4:
                        error = error.sum()

            self.preconditioner(psit_nX, residual_nX, out=psit2_nX)

            # Calculate projections
            wfs.pt_aiX.integrate(psit2_nX, out=P2_ani)

            # <psi2 | H | psi2>
            me(psit2_nX, psit2_nX, function=Ht)
            dH(P2_ani, out_ani=P3_ani)
            P2_ani.matrix.multiply(P3_ani, opb='C', symmetric=True, beta=1,
                                   out=M_nn)
            copy(H_NN.data[B:, B:])

            # <psi2 | H | psi>
            me(residual_nX, psit_nX)
            P3_ani.matrix.multiply(P_ani, opb='C', beta=1.0, out=M_nn)
            copy(H_NN.data[B:, :B])

            # <psi2 | S | psi2>
            me(psit2_nX, psit2_nX)
            dS(P2_ani, out_ani=P3_ani)
            P2_ani.matrix.multiply(P3_ani, opb='C', symmetric=True, beta=1,
                                   out=M_nn)
            copy(S_NN.data[B:, B:])

            # <psi2 | S | psi>
            me(psit2_nX, psit_nX)
            P3_ani.matrix.multiply(P_ani, opb='C', beta=1.0, out=M_nn)
            copy(S_NN.data[B:, :B])

            if is_domain_band_master:
                H_NN.data[:B, :B] = xp.diag(eig_N[:B])
                S_NN.data[:B, :B] = xp.eye(B)
                eig_N[:] = H_NN.eigh(S_NN)
                wfs._eig_n = as_xp(eig_N[:B], np)
            if domain_comm.rank == 0:
                band_comm.broadcast(wfs.eig_n, 0)
            domain_comm.broadcast(wfs.eig_n, 0)

            if domain_comm.rank == 0:
                if band_comm.rank == 0:
                    # M0_nn.data[:] = H_NN.data[:B, :B]
                    M0_nn.data[:] = H_NN.data[:B, :B].T
                M0_nn.redist(M_nn)
            domain_comm.broadcast(M_nn.data, 0)

            M_nn.multiply(psit_nX, opa='C', out=residual_nX)
            M_nn.multiply(P_ani, opa='C', out=P3_ani)

            if domain_comm.rank == 0:
                if band_comm.rank == 0:
                    # M0_nn.data[:] = H_NN.data[B:, :B]
                    M0_nn.data[:] = H_NN.data[:B, B:].T
                M0_nn.redist(M_nn)
            domain_comm.broadcast(M_nn.data, 0)

            M_nn.multiply(psit2_nX, opa='C', beta=1.0, out=residual_nX)
            M_nn.multiply(P2_ani, opa='C', beta=1.0, out=P3_ani)
            psit_nX.data[:] = residual_nX.data
            P_ani, P3_ani = P3_ani, P_ani
            wfs._P_ani = P_ani

            if i < self.niter - 1:
                Ht(psit_nX)
                calculate_residuals(residual_nX, dH, dS, wfs, P2_ani, P3_ani)

        return error


def calculate_residuals(residual_nX: DA,
                        dH: AAFunc,
                        dS: AAFunc,
                        wfs: PWFDWaveFunctions,
                        P1_ani: AA,
                        P2_ani: AA) -> None:

    eig_n = wfs.myeig_n
    xp = residual_nX.xp
    if xp is np:
        for r, e, p in zip(residual_nX.data, wfs.myeig_n, wfs.psit_nX.data):
            axpy(-e, p, r)
    else:
        eig_n = xp.asarray(eig_n)
        for r, e, p in zip(residual_nX.data, eig_n, wfs.psit_nX.data):
            r -= p * e

    dH(wfs.P_ani, P1_ani)
    if wfs.ncomponents < 4:
        subscripts = 'nI, n -> nI'
    else:
        subscripts = 'nsI, n -> nsI'
    if xp is np:
        np.einsum(subscripts, wfs.P_ani.data, eig_n, out=P2_ani.data)
    else:
        P2_ani.data[:] = xp.einsum(subscripts, wfs.P_ani.data, eig_n)
    dS(P2_ani, P2_ani)
    P1_ani.data -= P2_ani.data
    wfs.pt_aiX.add_to(residual_nX, P1_ani)


def calculate_weights(converge_bands: int | str,
                      ibzwfs: IBZWaveFunctions) -> list[Array1D | None]:
    """Calculate convergence weights for all eigenstates."""
    assert ibzwfs.band_comm.size == 1, 'not implemented!'
    weight_un = []
    nu = len(ibzwfs.wfs_qs) * ibzwfs.nspins
    nbands = ibzwfs.nbands

    if converge_bands == 'occupied':
        # Converge occupied bands:
        for wfs in ibzwfs:
            try:
                # Methfessel-Paxton or cold-smearing distributions can give
                # negative occupation numbers - so we take the absolute value:
                weight_n = np.abs(wfs.occ_n)
            except ValueError:
                # No eigenvalues yet:
                return [None] * nu
            weight_un.append(weight_n)
        return weight_un

    if converge_bands == 'all':
        return [np.ones(nbands)] * nu

    if isinstance(converge_bands, int):
        # Converge fixed number of bands:
        n = converge_bands

        for wfs in ibzwfs:
            weight_n = np.zeros(nbands)
            if n < 0:
                n += nbands
            weight_n[:n] = 1.0
            weight_un.append(weight_n)
        return weight_un

    # Converge states with energy up to CBM + delta:
    assert converge_bands.startswith('CBM+')
    delta = float(converge_bands[4:]) / Ha

    if ibzwfs.fermi_levels is None:
        return [None] * nu

    efermi = np.mean(ibzwfs.fermi_levels)

    # Find CBM:
    cbm = np.inf
    nocc_u = np.empty(nu, int)
    for u, wfs in enumerate(ibzwfs):
        n = (wfs.eig_n < efermi).sum()  # number of occupied bands
        nocc_u[u] = n
        if n < nbands:
            cbm = min(cbm, wfs.eig_n[n])

    # If all k-points don't have the same number of occupied bands,
    # then it's a metal:
    n0 = int(broadcast_float(float(nocc_u[0]), ibzwfs.kpt_comm))
    metal = bool(ibzwfs.kpt_comm.sum_scalar(float((nocc_u != n0).any())))
    if metal:
        cbm = efermi
    else:
        cbm = ibzwfs.kpt_comm.min_scalar(cbm)

    ecut = cbm + delta

    for wfs in ibzwfs:
        weight_n = (wfs.eig_n < ecut).astype(float)
        if wfs.eig_n[-1] < ecut:
            # We don't have enough bands!
            weight_n[:] = np.inf
        weight_un.append(weight_n)

    return weight_un



import pytest
from gpaw.utilities.elpa import LibElpa
import numpy as np
from gpaw.blacs import BlacsGrid
from gpaw.mpi import world

pytestmark = pytest.mark.skipif(not LibElpa.have_elpa(),
                                reason='not LibElpa.have_elpa()')


def test_libelpa():
    rng = np.random.RandomState(87878787)

    if world.size == 1:
        shape = 1, 1
    else:
        shape = world.size // 2, 2
    bg = BlacsGrid(world, *shape)

    M = 8
    blocksize = 2

    desc = bg.new_descriptor(M, M, blocksize, blocksize)
    sdesc = desc.as_serial()

    Aserial = sdesc.zeros()
    if world.rank == 0:
        Aserial[:] = rng.rand(*Aserial.shape)
        Aserial += Aserial.T.copy()
    A = desc.distribute_from_master(Aserial)
    C1 = desc.zeros()
    C2 = desc.zeros()
    eps1 = np.zeros(M)
    eps2 = np.zeros(M)

    elpa = LibElpa(desc)
    print(elpa)

    desc.diagonalize_dc(A.copy(), C1, eps1),

    elpa.diagonalize(A.copy(), C2, eps2)

    print(eps1)
    print(eps2)
    err = np.abs(eps1 - eps2).max()
    assert err < 1e-13, err



import pytest
import numpy as np

from ase import Atom, Atoms

from gpaw import GPAW, FermiDirac
from gpaw.test import gen
from gpaw.xas import XAS
import gpaw.mpi as mpi


@pytest.mark.skip(reason='TODO')
def test_si_nonortho():
    # Generate setup for oxygen with half a core-hole:
    gen('Si', name='hch1s', corehole=(1, 0, 0.5))

    a = 5.43095

    si_nonortho = Atoms([Atom('Si', (0, 0, 0)),
                         Atom('Si', (a / 4, a / 4, a / 4))],
                        cell=[(a / 2, a / 2, 0),
                              (a / 2, 0, a / 2),
                              (0, a / 2, a / 2)],
                        pbc=True)

    # calculation with full symmetry
    calc = GPAW(nbands=-10,
                h=0.25,
                kpts=(2, 2, 2),
                occupations=FermiDirac(width=0.05),
                setups={0: 'hch1s'})

    si_nonortho.calc = calc
    _ = si_nonortho.get_potential_energy()
    calc.write('si_nonortho_xas_sym.gpw')

    # calculation without any symmetry
    calc = GPAW(nbands=-10,
                h=0.25,
                kpts=(2, 2, 2),
                occupations=FermiDirac(width=0.05),
                setups={0: 'hch1s'},
                symmetry='off')

    si_nonortho.calc = calc
    _ = si_nonortho.get_potential_energy()
    calc.write('si_nonortho_xas_nosym.gpw')

    # restart from file
    calc1 = GPAW('si_nonortho_xas_sym.gpw')
    calc2 = GPAW('si_nonortho_xas_nosym.gpw')
    if mpi.size == 1:
        xas1 = XAS(calc1)
        x, y1 = xas1.get_spectra()
        xas2 = XAS(calc2)
        x2, y2 = xas2.get_spectra(E_in=x)

        assert (np.sum(abs(y1 - y2)[0, :500]**2) < 5e-9)
        assert (np.sum(abs(y1 - y2)[1, :500]**2) < 5e-9)
        assert (np.sum(abs(y1 - y2)[2, :500]**2) < 5e-9)



import numpy as np
from gpaw.mpi import world
from gpaw.utilities.partition import AtomPartition
from gpaw.arraydict import ArrayDict


def test_parallel_arraydict_redist():
    gen = np.random.RandomState(0)

    def shape(a):
        return (a, a // 2)  # Shapes: (0, 0), (1, 0), (2, 1), ...

    natoms = 33

    if world.size == 1:
        rank_a = np.zeros(natoms, int)
    else:
        # When on more than 2 cores, make sure that at least one core
        # (rank=0) has zero entries:
        lower = 0 if world.size == 2 else 1
        rank_a = gen.randint(lower, world.size, natoms)
    assert (rank_a < world.size).all()

    serial = AtomPartition(world, np.zeros(natoms, int))
    partition = AtomPartition(world, rank_a)
    even_partition = partition.as_even_partition()

    def check(atomdict, title):
        if world.rank == world.size // 2 or world.rank == 0:
            print('rank %d %s: %s' % (world.rank, title.rjust(10), atomdict))

        # Create a normal, "well-behaved" dict against which to test arraydict.
        ref = dict(atomdict)
        # print atomdict
        assert set(atomdict.keys()) == set(ref.keys())  # check keys()
        for a in atomdict:  # check __iter__, __getitem__
            assert ref[a] is atomdict[a]
        values = atomdict.values()
        for i, key in enumerate(atomdict):
            # AtomDict guarantees fixed ordering of keys.  Check that
            # values() ordering is consistent with loop ordering:
            assert values[i] is atomdict[key]

        items = list(atomdict.items())

        for i, (key, item) in enumerate(atomdict.items()):
            assert item is atomdict[key]
            assert item is ref[key]
            assert items[i][0] == key
            assert items[i][1] is item

        # Hopefully this should verify all the complicated stuff

    ad = ArrayDict(partition, shape, float)
    for key in ad:
        ad[key][:] = key
    array0 = ad.toarray()

    _ = dict(ad)
    check(ad, 'new')
    ad.redistribute(even_partition)
    array1 = ad.toarray()
    if world.rank > 1:
        assert array1.shape != array0.shape
    check(ad, 'even')
    ad.redistribute(serial)
    check(ad, 'serial')
    ad.redistribute(partition)
    check(ad, 'back')

    array2 = ad.toarray()
    assert (array0 == array2).all()



"""
Calculate the electron-phonon matrix in lithium using ALDA.

Tests whether the spin-paired and spin-polarised results are identical.
"""
import numpy as np
import pytest
from ase.build import bulk
from ase.parallel import world

from gpaw import GPAW
from gpaw.elph.electronphonon import ElectronPhononCoupling


@pytest.mark.elph
def test_elph_li(in_tmp_dir):
    # 2 atoms with one 1 valence electron each
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True)

    for spinpol in (0, 1):
        # Part 1: ground state calculation

        # sz:  1 orbital per atom
        # szp: 4 orbitals per atom
        # dz:  2 orbitals per atom -> fastest non-trivial case
        calc = GPAW(mode='lcao',
                    basis='sz(dzp)',
                    kpts={'size': (1, 1, 1), 'gamma': False},
                    # kpts=(2, 2, 2),
                    symmetry={'point_group': False},
                    convergence={'bands': 'nao'},
                    spinpol=spinpol,
                    parallel={'domain': 1},
                    txt='elph_li.txt'
                    )
        atoms.calc = calc

        # szp, k222: 4s/8s on 1 core
        atoms.get_potential_energy()

        # Part 2: compute displacements

        # szp, k222: 40s (80s) on 1 core
        # this step produces pickle files. do we need to clean them manually?
        if spinpol:
            name = 'elph_spinpol'
            sc_name = 'sc_spinpol'
        else:
            name = 'elph_spinpared'
            sc_name = 'sc_spinpaired'
        elph = ElectronPhononCoupling(atoms=atoms, calc=atoms.calc,
                                      supercell=(1, 1, 1), name=name,
                                      calculate_forces=False)
        elph.run()

        # Part 3: contruct coefficient matrix

        # szp, k222: 4s on 1 core
        elph.set_lcao_calculator(atoms.calc)
        elph.calculate_supercell_matrix(name=sc_name, include_pseudo=True)
        elph.load_supercell_matrix(name=sc_name)
        if spinpol:
            g_xsMM = elph.g_xsNNMM[:, :, 0, 0]
        else:
            g_xMM = elph.g_xsNNMM[:, 0, 0, 0]

        if world.rank == 0:  # others don't have g_xsNNMM
            # Part 4:  analyse matrix
            for s in range(spinpol + 1):
                for x in range(6):  # 2 atoms * 3 directions
                    # gMM is symmetric
                    assert (np.allclose(elph.g_xsNNMM[x, s, 0, 0],
                                        elph.g_xsNNMM[x, s, 0, 0].T))
                    # in this case both atoms and all displacements are
                    # equivalent
                    # all six gMM have same entries, but in different places
                    assert (abs(np.max(abs(elph.g_xsNNMM[x, s, 0, 0])) -
                                np.max(abs(elph.g_xsNNMM[0, 0, 0, 0]))) < 5e-5)
        # remove json cache
        elph.clean()

    if world.rank == 0:
        # Part 5: compare spin-paired and spin-polarised
        assert np.allclose(g_xsMM[:, 0], g_xsMM[:, 1])
        assert np.allclose(g_xMM, g_xsMM[:, 0])



import numpy as np
from ase.build import molecule
from ase.data.vdw import vdw_radii
from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.test import equal
from gpaw.solvation import (SolvationGPAW, EffectivePotentialCavity,
                            Power12Potential, LinearDielectric)

vdw_radii = vdw_radii.copy()


def test_solvation_vacuum():

    SKIP_REF_CALC = True

    energy_eps = 0.0005 / 8.
    forces_eps = 3e-2

    h = 0.3
    vac = 3.0
    u0 = .180
    T = 298.15
    vdw_radii[1] = 1.09

    def atomic_radii(atoms):
        return [vdw_radii[n] for n in atoms.numbers]

    atoms = Cluster(molecule('H2O'))
    atoms.minimal_box(vac, h)

    convergence = {
        'energy': energy_eps,
        'forces': forces_eps ** 2,  # Force error is squared
        'density': 10.,
        'eigenstates': 10.,
    }

    if not SKIP_REF_CALC:
        atoms.calc = GPAW(xc='LDA', h=h, convergence=convergence)
        Eref = atoms.get_potential_energy()
        print(Eref)
        Fref = atoms.get_forces()
        print(Fref)
    else:
        # setups: 0.9.11271, same settings as above
        Eref = -11.9932

        Fref = np.array(
            [[1.95122040e-12, -1.17770462e-12, -6.04993798e+00],
             [6.61270337e-14, 1.58227909e+00, 6.06605145e-02],
             [1.35947527e-13, -1.58227909e+00, 6.06605145e-02]])

    atoms.calc = SolvationGPAW(
        xc='LDA', h=h, convergence=convergence,
        cavity=EffectivePotentialCavity(
            effective_potential=Power12Potential(atomic_radii=atomic_radii,
                                                 u0=u0),
            temperature=T
        ),
        dielectric=LinearDielectric(epsinf=1.0),
    )
    Etest = atoms.get_potential_energy()
    Eeltest = atoms.calc.get_electrostatic_energy()
    Ftest = atoms.get_forces()
    equal(Etest, Eref, energy_eps * atoms.calc.get_number_of_electrons())
    equal(Ftest, Fref, forces_eps)
    equal(Eeltest, Etest)



"""Contains methods for calculating local LR-TDDFT kernels."""

from functools import partial
from pathlib import Path

import numpy as np

from gpaw.response import timer
from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.localft import (LocalFTCalculator,
                                   add_LDA_dens_fxc, add_LSDA_trans_fxc)


class FXCKernel:
    r"""Adiabatic local exchange-correlation kernel in a plane-wave basis.

    In real-space, the adiabatic local xc-kernel matrix is given by:

    K_xc^μν(r, r', t-t') = f_xc^μν(r) δ(r-r') δ(t-t')

    where the local xc kernel is given by:

                 ∂^2[ϵ_xc(n,m)n] |
    f_xc^μν(r) = ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾ |
                   ∂n^μ ∂n^ν     |n=n(r),m=m(r)

    In the plane-wave basis (and frequency domain),

                     1
    K_xc^μν(G, G') = ‾‾ f_xc^μν(G-G'),
                     V0

    where V0 is the cell volume.

    Because the set of unique reciprocal lattice vector differences
    dG = G-G' is much more compact than the full kernel matrix structure
    Kxc_GG', we store data internally in the dG representation and only
    unfold the data into the full kernel matrix when requested.
    """

    def __init__(self, fxc_dG, dG_K, GG_shape, volume):
        """Construct the fxc kernel."""
        assert np.prod(GG_shape) == len(dG_K),\
            "The K index should be a flattened (G,G') composite index'"

        self._fxc_dG = fxc_dG
        self._dG_K = dG_K
        self.GG_shape = GG_shape
        self.volume = volume

    def get_Kxc_GG(self):
        """Unfold the fxc(G-G') kernel into the Kxc_GG' kernel matrix."""
        # Kxc(G-G') = 1 / V0 * fxc(G-G')
        Kxc_dG = 1 / self.volume * self._fxc_dG

        # Unfold Kxc(G-G') to the kernel matrix structure Kxc_GG'
        Kxc_GG = Kxc_dG[self._dG_K].reshape(self.GG_shape)

        return Kxc_GG

    def save(self, path: Path):
        """Save the fxc kernel in a .npz file."""
        assert path.suffix == '.npz'
        with open(str(path), 'wb') as fd:
            np.savez(fd,
                     fxc_dG=self._fxc_dG,
                     dG_K=self._dG_K,
                     GG_shape=self.GG_shape,
                     volume=self.volume)

    @staticmethod
    def from_file(path: Path):
        """Construct an fxc kernel from a previous calculation."""
        assert path.suffix == '.npz'
        npzfile = np.load(path)

        args = [npzfile[key]
                for key in ['fxc_dG', 'dG_K', 'GG_shape', 'volume']]

        return FXCKernel(*args)


class AdiabaticFXCCalculator:
    """Calculator for adiabatic local exchange-correlation kernels."""

    def __init__(self, localft_calc: LocalFTCalculator):
        """Contruct the fxc calculator based on a local FT calculator."""
        self.localft_calc = localft_calc

        self.gs = localft_calc.gs
        self.context = localft_calc.context

    @staticmethod
    def from_rshe_parameters(*args, **kwargs):
        return AdiabaticFXCCalculator(
            LocalFTCalculator.from_rshe_parameters(*args, **kwargs))

    @timer('Calculate XC kernel')
    def __call__(self, fxc, spincomponent, qpd):
        """Calculate fxc(G-G'), which defines the kernel matrix Kxc_GG'.

        The fxc kernel is calculated for all unique dG = G-G' reciprocal
        lattice vectors and returned as an FXCKernel instance which can unfold
        itself to produce the full kernel matrix Kxc_GG'.
        """
        # Generate a large_qpd to encompass all G-G' in qpd
        large_ecut = 4 * qpd.ecut  # G = 1D grid of |G|^2/2 < ecut
        large_qpd = qpd.copy_with(ecut=large_ecut,
                                  gammacentered=True,
                                  gd=self.gs.finegd)
        
        # Calculate fxc(Q) on the large plane-wave grid (Q = large grid index)
        add_fxc = create_add_fxc(fxc, spincomponent)
        fxc_Q = self.localft_calc(large_qpd, add_fxc)

        # Create a mapping from the large plane-wave grid to an unfoldable mesh
        # of all unique dG = G-G' reciprocal lattice vector differences on the
        # qpd plane-wave representation
        GG_shape, dG_K, Q_dG = self.create_unfoldable_Q_dG_mapping(
            qpd, large_qpd)

        # Map the calculated kernel fxc(Q) onto the unfoldable grid of unique
        # reciprocal lattice vector differences fxc(dG)
        fxc_dG = fxc_Q[Q_dG]

        # Return the calculated kernel as an fxc kernel object
        fxc_kernel = FXCKernel(fxc_dG, dG_K, GG_shape, qpd.gd.volume)

        return fxc_kernel

    @timer('Create unfoldable Q_dG mapping')
    def create_unfoldable_Q_dG_mapping(self, qpd, large_qpd):
        """Create mapping from Q index to the kernel matrix indeces GG'.

        The mapping is split into two parts:
         * Mapping from the large plane-wave representation index Q (of
           large_qpd) to an index dG representing all unique reciprocal lattice
           vector differences (G-G') of the original plane-wave representation
           qpd
         * A mapping from the dG index to a flattened K = (G, G') composite
           kernel matrix index

        Lastly the kernel matrix shape GG_shape is returned so that an array in
        index K can easily be reshaped to a kernel matrix Kxc_GG'.
        """
        # Calculate all (G-G') reciprocal lattice vectors
        dG_GGv = calculate_dG_GGv(qpd)
        GG_shape = dG_GGv.shape[:2]

        # Reshape to composite K = (G, G') index
        dG_Kv = dG_GGv.reshape(-1, dG_GGv.shape[-1])

        # Find unique dG-vectors
        # We need tight control of the decimals to avoid precision artifacts
        dG_dGv, dG_K = np.unique(dG_Kv.round(decimals=6),
                                 return_inverse=True, axis=0)

        # Create the mapping from Q-index to dG-index
        Q_dG = self.create_Q_dG_map(large_qpd, dG_dGv)

        return GG_shape, dG_K, Q_dG

    @timer('Create Q_dG map')
    def create_Q_dG_map(self, large_qpd, dG_dGv):
        """Create mapping between (G-G') index dG and large_qpd index Q."""
        G_Qv = large_qpd.get_reciprocal_vectors(add_q=False)
        # Make sure to match the precision of dG_dGv
        G_Qv = G_Qv.round(decimals=6)

        # Distribute dG over world
        # This is necessary because the next step is to create a K_QdGv buffer
        # of which the norm is taken. When the number of plane-wave
        # coefficients is large, this step becomes a memory bottleneck, hence
        # the distribution.
        dGblocks = Blocks1D(self.context.comm, dG_dGv.shape[0])
        dG_mydGv = dG_dGv[dGblocks.myslice]

        # Determine Q index for each dG index
        diff_QmydG = np.linalg.norm(G_Qv[:, np.newaxis] - dG_mydGv[np.newaxis],
                                    axis=2)
        Q_mydG = np.argmin(diff_QmydG, axis=0)

        # Check that all the identified Q indices produce identical reciprocal
        # lattice vectors
        assert np.allclose(np.diagonal(diff_QmydG[Q_mydG]), 0.),\
            'Could not find a perfect matching reciprocal wave vector in '\
            'large_qpd for all dG_dGv'

        # Collect the global Q_dG map
        Q_dG = dGblocks.all_gather(Q_mydG)

        return Q_dG


def create_add_fxc(fxc, spincomponent):
    """Create an add_fxc function according to the requested functional and
    spin component."""
    assert fxc in ['ALDA_x', 'ALDA_X', 'ALDA']

    if spincomponent in ['00', 'uu', 'dd']:
        add_fxc = partial(add_LDA_dens_fxc, fxc=fxc)
    elif spincomponent in ['+-', '-+']:
        add_fxc = partial(add_LSDA_trans_fxc, fxc=fxc)
    else:
        raise ValueError(spincomponent)

    return add_fxc


def calculate_dG_GGv(qpd):
    """Calculate dG_GG' = (G-G') for the plane wave basis in qpd."""
    nG = qpd.ngmax
    G_Gv = qpd.get_reciprocal_vectors(add_q=False)

    dG_GGv = np.zeros((nG, nG, 3))
    for v in range(3):
        dG_GGv[:, :, v] = np.subtract.outer(G_Gv[:, v], G_Gv[:, v])

    return dG_GGv



from math import pi, sqrt

import numpy as np

from ase.units import Bohr

from gpaw.atom.all_electron import AllElectron
from gpaw.sphere.lebedev import weight_n, R_nv


# XXX why on earth is this necessary?
def get_scaled_positions(atoms, positions):
    """COPY PASTE FROM ASE! Get positions relative to unit cell.

    Atoms outside the unit cell will be wrapped into the cell in
    those directions with periodic boundary conditions so that the
    scaled coordinates are between zero and one."""

    scaled = np.linalg.solve(atoms.cell.T, positions.T).T
    for i in range(3):
        if atoms.pbc[i]:
            scaled[i] %= 1.0
    return scaled


class AllElectronPotential:
    def __init__(self, paw):
        self.paw = paw

    def write_spherical_ks_potentials(self, txt):
        f = open(txt, 'w')
        for a in self.paw.density.D_asp:
            r_g, vKS_g = self.get_spherical_ks_potential(a)
            setup = self.paw.density.setups[a]
            # Calculate also atomic LDA for reference
            g = AllElectron(setup.symbol,
                            xcname='LDA',
                            nofiles=True,
                            scalarrel=True,
                            txt=None)
            g.run()
            print(r_g[0], vKS_g[0], g.vr[0], 0.0, file=f)
            for r, vKS, vr in zip(r_g[1:], vKS_g[1:], g.vr[1:]):
                print(r, vKS, vr, (vKS - vr) / r, file=f)

        f.close()

    def grid_to_radial(self, a, gd, f_g):

        # Coordinates of an atom
        atom_c = self.paw.atoms.get_positions()[a]

        # Get xccorr for atom a
        setup = self.paw.density.setups[a]
        xccorr = setup.xc_correction

        radf_g = np.zeros(xccorr.ng)

        for w, p in zip(weight_n, R_nv):
            scaled_nc = []
            # Very inefficient loop
            for i, r in enumerate(xccorr.rgd.r_g):
                # Obtain the position of this integration quadrature point in
                # specified grid
                pos_c = atom_c + (r * Bohr) * p
                # And in scaled coordinates
                scaled_c = get_scaled_positions(self.paw.atoms, pos_c)
                scaled_nc.append(scaled_c)

            scaled_nc = np.array(scaled_nc)

            target_g = gd.interpolate_grid_points(scaled_nc, f_g)
            radf_g += w * target_g

        return radf_g

    def get_spherical_ks_potential(self, a):
        # self.paw.restore_state()

        print("XC:", self.paw.hamiltonian.xc.name)
        assert self.paw.hamiltonian.xc.type == 'LDA'

        # If the calculation is just loaded, density needs to be interpolated
        if self.paw.density.nt_sg is None:
            print("Interpolating density")
            self.paw.density.interpolate_pseudo_density()

        # Get xccorr for atom a
        setup = self.paw.density.setups[a]
        xccorr = setup.xc_correction

        # Get D_sp for atom a
        D_sp = self.paw.density.D_asp[a]

        # density a function of L and partial wave radial pair density
        # coefficient
        D_sLq = np.inner(D_sp, xccorr.B_pqL.T)

        # The 'spherical' spherical harmonic
        Y0 = 1.0 / sqrt(4 * pi)

        # Generate cartesian fine grid xc-potential
        print("Generate cartesian fine grid xc-potential")
        gd = self.paw.density.finegd
        vxct_sg = gd.zeros(1)
        xc = self.paw.hamiltonian.xc
        xc.calculate(gd, self.paw.density.nt_sg, vxct_sg)

        # ---------------------------------------------
        # The Electrostatic potential
        # ---------------------------------------------
        # V_ES(r) = Vt_ES(r) - Vt^a(r) + V^a(r), where
        # Vt^a = P[ nt^a(r) + \sum Q_L g_L(r) ]
        # V^a = P[ -Z\delta(r) + n^a(r) ]
        # P = Poisson solution
        # ---------------------------------------------

        print("Evaluating ES Potential...")
        # Make sure that the calculation has ES potential
        # TODO
        if self.paw.hamiltonian.vHt_g is None:
            raise "Converge tha Hartree potential first."

        # Interpolate the smooth electrostatic potential from fine grid to
        # radial grid
        radHt_g = self.grid_to_radial(a, gd, self.paw.hamiltonian.vHt_g)
        radHt_g *= xccorr.rgd.r_g

        print("D_sp", D_sp)

        # Calculate the difference in density and pseudo density
        dn_g = (Y0 * np.dot(D_sLq[0, 0], (xccorr.n_qg - xccorr.nt_qg)) +
                xccorr.nc_g - xccorr.nct_g)

        # Add the compensation charge contribution
        dn_g -= Y0 * self.paw.density.Q_aL[a][0] * setup.g_lg[0]

        # Calculate the Hartree potential for this
        vHr = xccorr.rgd.poisson(dn_g)

        # Add the core potential contribution
        vHr -= setup.Z

        radHt_g += vHr

        # --------------------------------------------
        # The XC potential
        # --------------------------------------------
        # V_xc = Vt_xc(r) - Vt_xc^a(r) + V_xc^a(r)
        # --------------------------------------------

        print("Evaluating xc potential")
        # Interpolate the smooth xc potential  from fine grid to radial grid
        radvxct_g = self.grid_to_radial(a, gd, vxct_sg[0])

        # Arrays for evaluating radial xc potential slice
        vxc_sg = np.zeros((len(D_sp), xccorr.ng))

        for n, Y_L in enumerate(xccorr.Y_nL):
            n_sLg = np.dot(D_sLq, xccorr.n_qg)
            n_sLg[:, 0] += sqrt(4 * pi) * xccorr.nc_g
            vxc_sg[:] = xc.calculate_radial(xccorr.rgd, n_sLg, Y_L)[1]
            radvxct_g += weight_n[n] * vxc_sg[0]
            nt_sLg = np.dot(D_sLq, xccorr.nt_qg)
            nt_sLg[:, 0] += sqrt(4 * pi) * xccorr.nct_g
            vxc_sg[:] = xc.calculate_radial(xccorr.rgd, nt_sLg, Y_L)[1]
            radvxct_g -= weight_n[n] * vxc_sg[0]

        radvks_g = radvxct_g * xccorr.rgd.r_g + radHt_g
        return (xccorr.rgd.r_g, radvks_g)


class CoreEigenvalues(AllElectronPotential):
    def get_core_eigenvalues(self, a, scalarrel=True):
        """Return the core eigenvalues by solving the radial schrodinger
        equation.

        Using AllElectron potential class, the spherically averaged Kohn--Sham
        potential is obtained around the spesified atom. The eigenstates for
        this potential are solved, the the resulting core states returned.
        Still experimental.

        """

        r, v_g = self.get_spherical_ks_potential(a)

        # Get xccorr for atom a
        setup = self.paw.density.setups[a]
        symbol = setup.symbol

        # Create AllElectron object for eigensolver
        atom = AllElectron(symbol, txt=None, scalarrel=scalarrel)
        # Calculate initial guess
        atom.run()

        # Set the potential
        atom.vr[:len(v_g)] = v_g
        # After the setups cutoff, arbitrary barrier is used
        atom.vr[len(v_g):] = 10.0

        # Solve the eigenstates
        atom.solve()

        # The display format is just copy paste from AllElectron class
        # TODO: Make it a method in AllElectron class, thus it can be called
        # directly
        def t(a):
            print(a)

        t('Calculated core eigenvalues of atom ' + str(a) + ':' + symbol)
        t('state      eigenvalue         ekin         rmax')
        t('-----------------------------------------------')
        for m, l, f, e, u in zip(atom.n_j, atom.l_j, atom.f_j, atom.e_j,
                                 atom.u_j):
            # Find kinetic energy:
            k = e - np.sum((
                np.where(abs(u) < 1e-160, 0, u)**2 *  # XXXNumeric!
                atom.vr * atom.dr)[1:] / atom.r[1:])

            # Find outermost maximum:
            g = atom.N - 4
            while u[g - 1] >= u[g]:
                g -= 1
            x = atom.r[g - 1:g + 2]
            y = u[g - 1:g + 2]
            A = np.transpose(np.array([x**i for i in range(3)]))
            c, b, a = np.linalg.solve(A, y)
            assert a < 0.0
            rmax = -0.5 * b / a

            s = 'spdf' [l]
            t('%d%s^%-4.1f: %12.6f %12.6f %12.3f' % (m, s, f, e, k, rmax))
        t('-----------------------------------------------')
        t('(units: Bohr and Hartree)')
        return atom.e_j



import numpy as np
import pytest
from ase import Atoms

from gpaw import GPAW, Davidson, FermiDirac


def test_fermilevel(in_tmp_dir):
    atoms = Atoms('He', pbc=True)
    atoms.center(vacuum=3)
    params = dict(nbands=1,
                  eigensolver=Davidson(6),
                  occupations=FermiDirac(0.0),
                  txt=None)

    atoms.calc = GPAW(**params)
    atoms.get_potential_energy()
    assert np.isinf(atoms.calc.get_fermi_level())

    params['nbands'] = 3
    params['convergence'] = {'bands': 2}
    atoms.calc = GPAW(**params)
    atoms.get_potential_energy()

    homo, lumo = atoms.calc.get_homo_lumo()
    assert homo == pytest.approx(-15.4473, abs=0.01)
    assert lumo == pytest.approx(-0.2566, abs=0.01)

    atoms.calc.write('test.gpw')
    print(GPAW('test.gpw').get_homo_lumo(), homo, lumo)
    assert np.all(GPAW('test.gpw').get_homo_lumo() == (homo, lumo))
    ef = atoms.calc.get_fermi_level()
    assert ef == pytest.approx(-7.85196, abs=0.01)



import numpy as np
import pytest
from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.mpi import world
from gpaw.tddft.spectrum import photoabsorption_spectrum
from gpaw.test import equal


@pytest.mark.rttddft
def test_lcaotddft_lcaotddft_vs_lrtddft2_rpa(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='sz(dzp)', mode='lcao', xc='oldLDA',
                convergence={'density': 1e-8},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', fxc='RPA', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick([0, 0, 1e-5])
    td_calc.propagate(30, 150)
    photoabsorption_spectrum('dm.dat', 'spec.dat',
                             e_max=10, width=0.5, delta_e=0.1)

    # LrTDDFT2 calculation
    calc = GPAW('gs.gpw', txt='lr.out')
    # It doesn't matter which fxc is here
    lr = LrTDDFT2('lr2', calc, fxc='LDA')
    lr.K_matrix.fxc_pre = 0.0  # Ignore fxc part
    lr.calculate()
    lr.get_spectrum('lr_spec.dat', 0, 10.1, 0.1, width=0.5)
    world.barrier()

    # Scale spectra due to slightly different definitions
    spec1_ej = np.loadtxt('spec.dat')
    data1_e = spec1_ej[:, 3]
    data1_e[1:] /= spec1_ej[1:, 0]
    spec2_ej = np.loadtxt('lr_spec.dat')
    E = lr.lr_transitions.get_transitions()[0][0]
    data2_e = spec2_ej[:, 5] / E

    # One can decrease the tolerance by decreasing the time step
    # and other parameters
    equal(data1_e, data2_e, 0.01)



import numpy as np
from gpaw.core import PlaneWaves
from gpaw.gpu import cupy as cp
from gpaw.gpu import is_hip
from gpaw.mpi import broadcast_float
from gpaw.new import zip
from gpaw.new.pot_calc import PotentialCalculator
from gpaw.new.pw.stress import calculate_stress
from gpaw.setup import Setups


class PlaneWavePotentialCalculator(PotentialCalculator):
    def __init__(self,
                 grid,
                 fine_grid,
                 pw: PlaneWaves,
                 fine_pw: PlaneWaves,
                 setups: Setups,
                 xc,
                 poisson_solver,
                 nct_ag,
                 nct_R,
                 soc=False,
                 xp=np):
        fracpos_ac = nct_ag.fracpos_ac
        atomdist = nct_ag.atomdist
        self.xp = xp
        super().__init__(xc, poisson_solver, setups, nct_R, fracpos_ac, soc)

        self.nct_ag = nct_ag
        self.vbar_ag = setups.create_local_potentials(
            pw, fracpos_ac, atomdist, xp)
        self.ghat_aLh = setups.create_compensation_charges(
            fine_pw, fracpos_ac, atomdist, xp)

        self.pw = pw
        self.fine_pw = fine_pw
        self.pw0 = pw.new(comm=None)  # not distributed

        self.h_g, self.g_r = fine_pw.map_indices(self.pw0)
        if xp is cp:
            self.h_g = cp.asarray(self.h_g)
            self.g_r = [cp.asarray(g) for g in self.g_r]

        # There is a bug in HIP cupyx.scipy.rfftn:
        self.xp0 = cp if xp is cp and not is_hip else np

        self.fftplan = grid.fft_plans(xp=self.xp0)
        self.fftplan2 = fine_grid.fft_plans(xp=self.xp0)

        self.grid = grid
        self.fine_grid = fine_grid

        self.vbar_g = pw.zeros(xp=xp)
        self.vbar_ag.add_to(self.vbar_g)
        self.vbar0_g = self.vbar_g.gather()

        self._nt_g = None
        self._vt_g = None

        self.e_stress = np.nan

    def calculate_charges(self, vHt_h):
        return self.ghat_aLh.integrate(vHt_h)

    def calculate_non_selfconsistent_exc(self, nt_sR, xc):
        nt_sr, _, _ = self._interpolate_density(nt_sR)
        vxct_sr = nt_sr.desc.empty(nt_sr.dims)
        e_xc = xc.calculate(nt_sr, vxct_sr)
        return e_xc

    def _interpolate_density(self, nt_sR):
        nt_sr = self.fine_grid.empty(nt_sR.dims, xp=self.xp0)
        pw = self.vbar_g.desc

        if pw.comm.rank == 0:
            indices = self.xp0.asarray(self.pw0.indices(self.fftplan.shape))
            nt0_g = self.pw0.zeros(xp=self.xp)
        else:
            nt0_g = None

        ndensities = nt_sR.dims[0] % 3
        for spin, (nt_R, nt_r) in enumerate(zip(nt_sR, nt_sr)):
            nt_R.to_xp(self.xp0).interpolate(
                self.fftplan, self.fftplan2, out=nt_r)
            if spin < ndensities and pw.comm.rank == 0:
                nt0_g.data += self.xp.asarray(
                    self.fftplan.tmp_Q.ravel()[indices])

        return nt_sr.to_xp(self.xp), pw, nt0_g

    def calculate_pseudo_potential(self, density, vHt_h):
        nt_sr, pw, nt0_g = self._interpolate_density(density.nt_sR)

        vxct_sr = nt_sr.desc.empty(density.nt_sR.dims, xp=self.xp)
        e_xc = self.xc.calculate(nt_sr, vxct_sr)

        if pw.comm.rank == 0:
            nt0_g.data *= 1 / np.prod(density.nt_sR.desc.size_c)
            e_zero = self.vbar0_g.integrate(nt0_g)
        else:
            e_zero = 0.0
        e_zero = broadcast_float(float(e_zero), pw.comm)

        if vHt_h is None:
            vHt_h = self.ghat_aLh.pw.zeros(xp=self.xp)

        charge_h = vHt_h.desc.zeros(xp=self.xp)
        coef_aL = density.calculate_compensation_charge_coefficients()
        self.ghat_aLh.add_to(charge_h, coef_aL)

        if pw.comm.rank == 0:
            for rank, g in enumerate(self.g_r):
                if rank == 0:
                    charge_h.data[self.h_g] += nt0_g.data[g]
                else:
                    pw.comm.send(nt0_g.data[g], rank)
        else:
            data = self.xp.empty(len(self.h_g), complex)
            pw.comm.receive(data, 0)
            charge_h.data[self.h_g] += data

        # background charge ???

        e_coulomb = self.poisson_solver.solve(vHt_h, charge_h)

        if pw.comm.rank == 0:
            vt0_g = self.vbar0_g.copy()
            for rank, g in enumerate(self.g_r):
                if rank == 0:
                    vt0_g.data[g] += vHt_h.data[self.h_g]
                else:
                    data = self.xp.empty(len(g), complex)
                    pw.comm.receive(data, rank)
                    vt0_g.data[g] += data
            vt0_R = vt0_g.to_xp(self.xp0).ifft(
                plan=self.fftplan,
                grid=density.nt_sR.desc.new(comm=None)).to_xp(self.xp)
        else:
            pw.comm.send(vHt_h.data[self.h_g], 0)

        vt_sR = density.nt_sR.new()
        vt_sR[0].scatter_from(vt0_R if pw.comm.rank == 0 else None)
        if density.ndensities == 2:
            vt_sR.data[1] = vt_sR.data[0]
        vt_sR.data[density.ndensities:] = 0.0

        e_kinetic = self._restrict(vxct_sr, vt_sR, density)

        e_external = 0.0

        self.e_stress = e_coulomb + e_zero

        self._reset()

        return {'kinetic': e_kinetic,
                'coulomb': e_coulomb,
                'zero': e_zero,
                'xc': e_xc,
                'external': e_external}, vt_sR, vHt_h

    def _restrict(self, vxct_sr, vt_sR, density=None):
        vtmp_R = vt_sR.desc.empty(xp=self.xp0)
        e_kinetic = 0.0
        for spin, (vt_R, vxct_r) in enumerate(zip(vt_sR, vxct_sr)):
            vxct_r.to_xp(self.xp0).fft_restrict(
                self.fftplan2, self.fftplan, out=vtmp_R)
            vt_R.data += vtmp_R.to_xp(self.xp).data
            if density:
                e_kinetic -= vt_R.integrate(density.nt_sR[spin])
                if spin < density.ndensities:
                    e_kinetic += vt_R.integrate(self.nct_R)
        return float(e_kinetic)

    def restrict(self, vt_sr):
        vt_sR = self.grid.empty(vt_sr.dims, xp=self.xp0)
        for vt_R, vt_r in zip(vt_sR, vt_sr):
            vt_r.to_xp(self.xp0).fft_restrict(
                self.fftplan2, self.fftplan, out=vt_R)
        return vt_sR

    def _move(self, fracpos_ac, atomdist, ndensities):
        self.ghat_aLh.move(fracpos_ac, atomdist)
        self.vbar_ag.move(fracpos_ac, atomdist)
        self.vbar_g.data[:] = 0.0
        self.vbar_ag.add_to(self.vbar_g)
        self.vbar0_g = self.vbar_g.gather()
        self.nct_ag.move(fracpos_ac, atomdist)
        self.nct_ag.to_uniform_grid(out=self.nct_R, scale=1.0 / ndensities)
        self._reset()

    def _reset(self):
        self._vt_g = None
        self._nt_g = None

    def _force_stress_helper(self, state):
        if self._vt_g is None:
            density = state.density
            potential = state.potential
            nt_R = density.nt_sR[0]
            vt_R = potential.vt_sR[0]
            if density.ndensities > 1:
                nt_R = nt_R.desc.empty(xp=self.xp)
                nt_R.data[:] = density.nt_sR.data[:density.ndensities].sum(
                    axis=0)
                vt_R = vt_R.desc.empty(xp=self.xp)
                vt_R.data[:] = (
                    potential.vt_sR.data[:density.ndensities].sum(axis=0) /
                    density.ndensities)
            self._vt_g = vt_R.to_xp(self.xp0).fft(self.fftplan,
                                                  pw=self.pw).to_xp(self.xp)
            self._nt_g = nt_R.to_xp(self.xp0).fft(self.fftplan,
                                                  pw=self.pw).to_xp(self.xp)
        return self._vt_g, self._nt_g

    def force_contributions(self, state):
        vt_g, nt_g = self._force_stress_helper(state)

        return (self.ghat_aLh.derivative(state.vHt_x),
                self.nct_ag.derivative(vt_g),
                self.vbar_ag.derivative(nt_g))

    def stress(self, state):
        vt_g, nt_g = self._force_stress_helper(state)
        return calculate_stress(self, state, vt_g, nt_g)



def create_wave_function_mode(name, **kwargs):
    if name not in ['fd', 'pw', 'lcao']:
        raise ValueError('Unknown wave function mode: ' + name)

    from gpaw.wavefunctions.fd import FD
    from gpaw import PW
    from gpaw.wavefunctions.lcao import LCAO
    return {'fd': FD, 'pw': PW, 'lcao': LCAO}[name](**kwargs)


class Mode:
    def __init__(self, force_complex_dtype=False):
        self.force_complex_dtype = force_complex_dtype

    def todict(self):
        dct = {'name': self.name}
        if self.force_complex_dtype:
            dct['force_complex_dtype'] = True
        return dct



"""Tetrahedron method for Brillouin-zone integrations.

See::

    Improved tetrahedron method for Brillouin-zone integrations.

    Peter E. Blöchl, O. Jepsen, and O. K. Andersen
    Phys. Rev. B 49, 16223 – Published 15 June 1994

    DOI:https://doi.org/10.1103/PhysRevB.49.16223
"""

from math import nan
from typing import List, Tuple, cast
import numpy as np
from scipy.spatial import Delaunay

from gpaw.occupations import (ZeroWidth, findroot, collect_eigelvalues,
                              distribute_occupation_numbers,
                              OccupationNumberCalculator, ParallelLayout)
from gpaw.mpi import broadcast_float
from gpaw.typing import Array1D, Array2D, Array3D, ArrayLike1D, ArrayLike2D


def bja1(e1: Array1D, e2: Array1D, e3: Array1D, e4: Array1D
         ) -> Tuple[float, Array1D]:
    """Eq. (A2) and (C2) from Blöchl, Jepsen and Andersen."""
    x = 1.0 / ((e2 - e1) * (e3 - e1) * (e4 - e1))
    return (-(e1**3).dot(x),
            3 * e1**2 * x)


def bja2(e1: Array1D, e2: Array1D, e3: Array1D, e4: Array1D
         ) -> Tuple[float, Array1D]:
    """Eq. (A3) and (C3) from Blöchl, Jepsen and Andersen."""
    x = 1.0 / ((e3 - e1) * (e4 - e1))
    y = (e3 - e1 + e4 - e2) / ((e3 - e2) * (e4 - e2))
    return (x.dot((e2 - e1)**2
                  - 3 * (e2 - e1) * e2
                  + 3 * e2**2
                  + y * e2**3),
            x * (3 * (e2 - e1)
                 - 6 * e2
                 - 3 * y * e2**2))


def bja3(e1: Array1D, e2: Array1D, e3: Array1D, e4: Array1D
         ) -> Tuple[float, Array1D]:
    """Eq. (A4) and (C4) from Blöchl, Jepsen and Andersen."""
    x = 1.0 / ((e4 - e1) * (e4 - e2) * (e4 - e3))
    return (len(e1) - x.dot(e4**3),
            3 * x * e4**2)


def bja1b(e1: Array1D, e2: Array1D, e3: Array1D, e4: Array1D) -> Array2D:
    """Eq. (B2)-(B5) from Blöchl, Jepsen and Andersen."""
    C = -0.25 * e1**3 / ((e2 - e1) * (e3 - e1) * (e4 - e1))
    w2 = -C * e1 / (e2 - e1)
    w3 = -C * e1 / (e3 - e1)
    w4 = -C * e1 / (e4 - e1)
    w1 = 4 * C - w2 - w3 - w4
    return np.array([w1, w2, w3, w4])


def bja2b(e1: Array1D, e2: Array1D, e3: Array1D, e4: Array1D) -> Array2D:
    """Eq. (B7)-(B10) from Blöchl, Jepsen and Andersen."""
    C1 = 0.25 * e1**2 / ((e4 - e1) * (e3 - e1))
    C2 = 0.25 * e1 * e2 * e3 / ((e4 - e1) * (e3 - e2) * (e3 - e1))
    C3 = 0.25 * e2**2 * e4 / ((e4 - e2) * (e3 - e2) * (e4 - e1))
    w1 = C1 + (C1 + C2) * e3 / (e3 - e1) + (C1 + C2 + C3) * e4 / (e4 - e1)
    w2 = C1 + C2 + C3 + (C2 + C3) * e3 / (e3 - e2) + C3 * e4 / (e4 - e2)
    w3 = (C1 + C2) * e1 / (e1 - e3) - (C2 + C3) * e2 / (e3 - e2)
    w4 = (C1 + C2 + C3) * e1 / (e1 - e4) + C3 * e2 / (e2 - e4)
    return np.array([w1, w2, w3, w4])


def bja3b(e1: Array1D, e2: Array1D, e3: Array1D, e4: Array1D) -> Array2D:
    """Eq. (B14)-(B17) from Blöchl, Jepsen and Andersen."""
    C = 0.25 * e4**3 / ((e4 - e1) * (e4 - e2) * (e4 - e3))
    w1 = 0.25 - C * e4 / (e4 - e1)
    w2 = 0.25 - C * e4 / (e4 - e2)
    w3 = 0.25 - C * e4 / (e4 - e3)
    w4 = 1.0 - 4 * C - w1 - w2 - w3
    return np.array([w1, w2, w3, w4])


def triangulate_submesh(rcell_cv: Array2D) -> Array3D:
    """Find the 6 tetrahedra."""
    ABC_sc = np.array([[A, B, C]
                       for A in [0, 1] for B in [0, 1] for C in [0, 1]])
    dt = Delaunay(ABC_sc.dot(rcell_cv))
    s_tq = dt.simplices
    ABC_tqc = ABC_sc[s_tq]

    # Remove zero-volume slivers:
    ABC_tqc = ABC_tqc[np.linalg.det(ABC_tqc[:, 1:] - ABC_tqc[:, :1]) != 0]

    assert ABC_tqc.shape == (6, 4, 3)
    return ABC_tqc


def triangulate_everything(size_c: Array1D,
                           ABC_tqc: Array3D,
                           i_k: Array1D) -> Array3D:
    """Triangulate the whole BZ.

    Returns i_ktq ndarray mapping:

    * k: BZ k-point index (0, 1, ...,  nbzk - 1)
    * t: tetrahedron index (0, 1, ..., 5)
    * q: tetrahedron corner index (0, 1, 2, 3)

    to i: IBZ k-point index (0, 1, ...,  nibzk - 1).
    """
    nbzk = cast(int, size_c.prod())
    ABC_ck = np.unravel_index(np.arange(nbzk), size_c)
    ABC_tqck = ABC_tqc[..., np.newaxis] + ABC_ck
    ABC_cktq = np.transpose(ABC_tqck, (2, 3, 0, 1))
    k_ktq = np.ravel_multi_index(
        ABC_cktq.reshape((3, nbzk * 6 * 4)),
        size_c,
        mode='wrap').reshape((nbzk, 6, 4))  # type: ignore
    i_ktq = i_k[k_ktq]
    return i_ktq


class TetrahedronMethod(OccupationNumberCalculator):
    name = 'tetrahedron-method'
    extrapolate_factor = 0.0

    def __init__(self,
                 rcell: ArrayLike2D,
                 size: ArrayLike1D,
                 improved=False,
                 bz2ibzmap: ArrayLike1D = None,
                 parallel_layout: ParallelLayout = None):
        """Tetrahedron method for calculating occupation numbers.

        The reciprocal cell, *rcell*, can be given in arbitrary units
        (only the shape matters) and *size* is the size of the
        Monkhorst-Pack grid.  If k-points have been symmetry-reduced
        the *bz2ibzmap* parameter  mapping BZ k-point indizes to
        IBZ k-point indices must be given.
        """

        OccupationNumberCalculator.__init__(self, parallel_layout)

        self.rcell_cv = np.asarray(rcell)
        self.size_c = np.asarray(size)
        self.improved = improved

        nbzk = self.size_c.prod()

        if bz2ibzmap is None:
            bz2ibzmap = np.arange(nbzk)

        self.i_k = np.asarray(bz2ibzmap)

        assert self.size_c.shape == (3,)
        assert self.rcell_cv.shape == (3, 3)
        assert self.i_k.shape == (nbzk,)

        ABC_tqc = triangulate_submesh(
            self.rcell_cv / self.size_c[:, np.newaxis])

        self.i_ktq = triangulate_everything(self.size_c, ABC_tqc, self.i_k)

        self.nibzkpts = self.i_k.max() + 1

    def __repr__(self):
        return (
            'TetrahedronMethod('
            f'rcell={self.rcell_cv.tolist()}, '
            f'size={self.size_c.tolist()}, '
            f'bz2ibzmap={self.i_k.tolist()}, '
            'parallel_layout=<'
            f'{self.bd.comm.size}x{self.kpt_comm.size}x{self.domain_comm.size}'
            '>)')

    def copy(self,
             parallel_layout: ParallelLayout = None,
             bz2ibzmap: List[int] = None
             ) -> OccupationNumberCalculator:
        return TetrahedronMethod(
            self.rcell_cv,
            self.size_c,
            self.improved,
            self.i_k if bz2ibzmap is None else bz2ibzmap,
            parallel_layout or self.parallel_layout)

    def _calculate(self,
                   nelectrons,
                   eig_qn,
                   weight_q,
                   f_qn,
                   fermi_level_guess=nan) -> Tuple[float, float]:
        if np.isnan(fermi_level_guess):
            zero = ZeroWidth(self.parallel_layout)
            fermi_level_guess, _ = zero._calculate(
                nelectrons, eig_qn, weight_q, f_qn)
            if np.isinf(fermi_level_guess):
                return fermi_level_guess, 0.0

        x = fermi_level_guess

        eig_in, weight_i, nkpts_r = collect_eigelvalues(eig_qn, weight_q,
                                                        self.bd, self.kpt_comm)

        if eig_in.size != 0:
            if len(eig_in) == self.nibzkpts:
                nspins = 1
            else:
                nspins = 2
                assert len(eig_in) == 2 * self.nibzkpts

            def func(x, eig_in=eig_in):
                """Return excess electrons and derivative."""
                if nspins == 1:
                    n, dn = count(x, eig_in, self.i_ktq)
                else:
                    n1, dn1 = count(x, eig_in[::2], self.i_ktq)
                    n2, dn2 = count(x, eig_in[1::2], self.i_ktq)
                    n = n1 + n2
                    dn = dn1 + dn2
                return n - nelectrons, dn

            fermi_level, niter = findroot(func, x)

            def w(de_in):
                return weights(de_in, self.i_ktq, self.improved)

            if nspins == 1:
                f_in = w(eig_in - fermi_level)
            else:
                f_in = np.zeros_like(eig_in)
                f_in[::2] = w(eig_in[::2] - fermi_level)
                f_in[1::2] = w(eig_in[1::2] - fermi_level)

            f_in *= 1 / (weight_i[:, np.newaxis] * len(self.i_k))
        else:
            f_in = None
            fermi_level = nan

        distribute_occupation_numbers(f_in, f_qn, nkpts_r,
                                      self.bd, self.kpt_comm)

        if self.kpt_comm.rank == 0:
            fermi_level = broadcast_float(fermi_level, self.bd.comm)
        fermi_level = broadcast_float(fermi_level, self.kpt_comm)

        return fermi_level, 0.0


def count(fermi_level: float,
          eig_in: Array2D,
          i_ktq: Array3D) -> Tuple[float, float]:
    """Count electrons.

    Return number of electrons and derivative with respect to fermi level.
    """
    eig_in = eig_in - fermi_level
    nocc_i = (eig_in < 0.0).sum(axis=1)
    n1 = nocc_i.min()
    n2 = nocc_i.max()

    ne = n1
    dnedef = 0.0

    if n1 == n2:
        return ne, dnedef

    ntetra = 6 * i_ktq.shape[0]
    eig_Tq = eig_in[i_ktq, n1:n2].transpose((0, 1, 3, 2)).reshape(
        (ntetra * (n2 - n1), 4))
    eig_Tq.sort(axis=1)

    eig_Tq = eig_Tq[eig_Tq[:, 0] < 0.0]

    mask1_T = eig_Tq[:, 1] > 0.0
    mask2_T = ~mask1_T & (eig_Tq[:, 2] > 0.0)
    mask3_T = ~mask1_T & ~mask2_T & (eig_Tq[:, 3] > 0.0)

    for mask_T, bjaa in [(mask1_T, bja1), (mask2_T, bja2), (mask3_T, bja3)]:
        n, dn_T = bjaa(*eig_Tq[mask_T].T)
        ne += n / ntetra
        dnedef += dn_T.sum() / ntetra  # type: ignore

    mask4_T = ~mask1_T & ~mask2_T & ~mask3_T
    ne += mask4_T.sum() / ntetra

    return ne, dnedef


def weights(eig_in: Array2D, i_ktq: Array3D, improved=False) -> Array2D:
    """Calculate occupation numbers."""
    nocc_i = (eig_in < 0.0).sum(axis=1)
    n1 = nocc_i.min()
    n2 = nocc_i.max()

    f_in = np.zeros_like(eig_in)

    for i in i_ktq[:, 0, 0]:
        f_in[i, :n1] += 6.0

    if n1 == n2:
        return f_in / 6

    ntetra = 6 * i_ktq.shape[0]
    eig_Tq = eig_in[i_ktq, n1:n2].transpose((0, 1, 3, 2)).reshape(
        (ntetra * (n2 - n1), 4))
    q_Tq = eig_Tq.argsort(axis=1)
    eig_Tq = np.take_along_axis(eig_Tq, q_Tq, 1)
    f_Tq = np.zeros_like(eig_Tq)

    mask0_T = eig_Tq[:, 0] > 0.0
    mask1_T = ~mask0_T & (eig_Tq[:, 1] > 0.0)
    mask2_T = ~mask0_T & ~mask1_T & (eig_Tq[:, 2] > 0.0)
    mask3_T = ~mask0_T & ~mask1_T & ~mask2_T & (eig_Tq[:, 3] > 0.0)

    for mask_T, bjab in [(mask1_T, bja1b), (mask2_T, bja2b), (mask3_T, bja3b)]:
        w_qT = bjab(*eig_Tq[mask_T].T)
        f_Tq[mask_T] += w_qT.T

    if improved:
        for mask_T, bja in [(mask1_T, bja1),
                            (mask2_T, bja2),
                            (mask3_T, bja3)]:
            e_Tq = eig_Tq[mask_T]
            _, d_T = bja(*e_Tq.T)
            f_Tq[mask_T] += (d_T * (e_Tq.sum(1) - 4 * e_Tq.T)).T / 40

    mask4_T = ~mask0_T & ~mask1_T & ~mask2_T & ~mask3_T
    f_Tq[mask4_T] += 0.25

    ktn_T = np.array(np.unravel_index(np.arange(len(eig_Tq)),
                                      (len(i_ktq), 6, n2 - n1))).T
    for f_q, q_q, (k, t, n) in zip(f_Tq, q_Tq, ktn_T):
        for q, f in zip(q_q, f_q):
            f_in[i_ktq[k, t, q], n1 + n] += f

    f_in *= 1 / 6

    return f_in



from __future__ import annotations
from math import sqrt, pi
import numpy as np
from ase.units import Bohr
from gpaw.typing import Vector
from gpaw.core.atom_centered_functions import AtomArraysLayout
from gpaw.utilities import unpack2, unpack
from gpaw.core.atom_arrays import AtomArrays
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.gpu import as_xp
from gpaw.new import zip
from gpaw.core.plane_waves import PlaneWaves


class Density:
    def __init__(self,
                 nt_sR: UniformGridFunctions,
                 D_asii: AtomArrays,
                 charge: float,
                 delta_aiiL: list,
                 delta0_a,
                 N0_aii,
                 l_aj):
        self.nt_sR = nt_sR
        self.D_asii = D_asii
        self.delta_aiiL = delta_aiiL
        self.delta0_a = delta0_a
        self.N0_aii = N0_aii
        self.l_aj = l_aj
        self.charge = charge

        self.ncomponents = nt_sR.dims[0]
        self.ndensities = self.ncomponents % 3
        self.collinear = self.ncomponents != 4
        self.natoms = len(delta0_a)

    def __repr__(self):
        return f'Density({self.nt_sR}, {self.D_asii}, charge={self.charge})'

    def __str__(self) -> str:
        return (f'density:\n'
                f'  components: {self.ncomponents}\n'
                f'  grid points: {self.nt_sR.desc.size}\n'
                f'  charge: {self.charge}  # |e|\n')

    def new(self, grid):
        old_grid = self.nt_sR.desc
        nt_sR = grid.empty(self.ncomponents, xp=self.nt_sR.xp)
        ecut = 0.999 * min(grid.ecut_max(), old_grid.ecut_max())
        pw = PlaneWaves(ecut=ecut, cell=old_grid.cell, comm=grid.comm)
        for nt_R, old_nt_R in zip(nt_sR, self.nt_sR):
            old_nt_R.fft(pw=pw).ifft(out=nt_R)

        return Density(nt_sR,
                       self.D_asii,
                       self.charge,
                       self.delta_aiiL,
                       self.delta0_a,
                       self.N0_aii,
                       self.l_aj)

    def calculate_compensation_charge_coefficients(self) -> AtomArrays:
        xp = self.D_asii.layout.xp
        ccc_aL = AtomArraysLayout(
            [delta_iiL.shape[2] for delta_iiL in self.delta_aiiL],
            atomdist=self.D_asii.layout.atomdist,
            xp=xp).empty()

        for a, D_sii in self.D_asii.items():
            Q_L = xp.einsum('sij, ijL -> L',
                            D_sii[:self.ndensities], self.delta_aiiL[a])
            Q_L[0] += self.delta0_a[a]
            ccc_aL[a] = Q_L

        return ccc_aL

    def normalize(self):
        comp_charge = 0.0
        xp = self.D_asii.layout.xp
        for a, D_sii in self.D_asii.items():
            comp_charge += xp.einsum('sij, ij ->',
                                     D_sii[:self.ndensities],
                                     self.delta_aiiL[a][:, :, 0])
            comp_charge += self.delta0_a[a]
        # comp_charge could be cupy.ndarray:
        comp_charge = float(comp_charge) * sqrt(4 * pi)
        comp_charge = self.nt_sR.desc.comm.sum(comp_charge)
        charge = comp_charge + self.charge
        pseudo_charge = self.nt_sR[:self.ndensities].integrate().sum()
        if pseudo_charge != 0.0:
            x = -charge / pseudo_charge
            self.nt_sR.data *= x

    def update(self, nct_R, ibzwfs):
        self.nt_sR.data[:] = 0.0
        self.D_asii.data[:] = 0.0
        ibzwfs.add_to_density(self.nt_sR, self.D_asii)
        self.nt_sR.data[:self.ndensities] += nct_R.data
        self.symmetrize(ibzwfs.ibz.symmetries)

    def symmetrize(self, symmetries):
        self.nt_sR.symmetrize(symmetries.rotation_scc,
                              symmetries.translation_sc)
        xp = self.nt_sR.xp
        D_asii = self.D_asii.gather(broadcast=True, copy=True)
        for a1, D_sii in self.D_asii.items():
            D_sii[:] = 0.0
            rotation_sii = symmetries.rotations(self.l_aj[a1], xp)
            for a2, rotation_ii in zip(symmetries.a_sa[:, a1],
                                       rotation_sii):
                D_sii += xp.einsum('ij, sjk, lk -> sil',
                                   rotation_ii, D_asii[a2], rotation_ii)
        self.D_asii.data *= 1.0 / len(symmetries)

    def move(self, delta_nct_R):
        self.nt_sR.data[:self.ndensities] += delta_nct_R.data

    @classmethod
    def from_superposition(cls,
                           grid,
                           nct_R,
                           atomdist,
                           setups,
                           basis_set,
                           magmom_av,
                           ncomponents,
                           charge=0.0,
                           hund=False):
        f_asi = {a: atomic_occupation_numbers(setup,
                                              magmom_v,
                                              ncomponents,
                                              hund,
                                              charge / len(setups))
                 for a, (setup, magmom_v) in enumerate(zip(setups, magmom_av))}

        nt_sR = nct_R.desc.zeros(ncomponents)
        basis_set.add_to_density(nt_sR.data, f_asi)
        ndensities = ncomponents % 3
        nt_sR.data[:ndensities] += nct_R.to_xp(np).data

        atom_array_layout = AtomArraysLayout([(setup.ni, setup.ni)
                                              for setup in setups],
                                             atomdist=atomdist)
        D_asii = atom_array_layout.empty(ncomponents)
        for a, D_sii in D_asii.items():
            D_sii[:] = unpack2(setups[a].initialize_density_matrix(f_asi[a]))

        xp = nct_R.xp
        return cls.from_data_and_setups(nt_sR.to_xp(xp),
                                        D_asii.to_xp(xp),
                                        charge,
                                        setups)

    @classmethod
    def from_data_and_setups(cls,
                             nt_sR,
                             D_asii,
                             charge,
                             setups):
        xp = nt_sR.xp
        return cls(nt_sR,
                   D_asii,
                   charge,
                   [xp.asarray(setup.Delta_iiL) for setup in setups],
                   [setup.Delta0 for setup in setups],
                   [unpack(setup.N0_p) for setup in setups],
                   [setup.l_j for setup in setups])

    def calculate_dipole_moment(self, fracpos_ac):
        dip_v = np.zeros(3)
        ccc_aL = self.calculate_compensation_charge_coefficients()
        ccc_aL = ccc_aL.to_cpu()
        pos_av = fracpos_ac @ self.nt_sR.desc.cell_cv
        for a, ccc_L in ccc_aL.items():
            c = ccc_L[0]
            dip_v -= c * (4 * pi)**0.5 * pos_av[a]
            if len(ccc_L) > 1:
                y, z, x = ccc_L[1:4]
                dip_v -= np.array([x, y, z]) * (4 * pi / 3)**0.5
        self.nt_sR.desc.comm.sum(dip_v)
        for nt_R in self.nt_sR:
            dip_v -= as_xp(nt_R.moment(), np)
        return dip_v

    def calculate_magnetic_moments(self):
        magmom_av = np.zeros((self.natoms, 3))
        magmom_v = np.zeros(3)
        domain_comm = self.nt_sR.desc.comm

        if self.ncomponents == 2:
            for a, D_sii in self.D_asii.items():
                M_ii = D_sii[0] - D_sii[1]
                magmom_av[a, 2] = np.einsum('ij, ij ->', M_ii, self.N0_aii[a])
                magmom_v[2] += (np.einsum('ij, ij ->', M_ii,
                                          self.delta_aiiL[a][:, :, 0]) *
                                sqrt(4 * pi))
            domain_comm.sum(magmom_av)
            domain_comm.sum(magmom_v)

            M_s = self.nt_sR.integrate()
            magmom_v[2] += M_s[0] - M_s[1]

        elif self.ncomponents == 4:
            for a, D_sii in self.D_asii.items():
                M_vii = D_sii[1:4]
                magmom_av[a] = np.einsum('vij, ij -> v',
                                         M_vii, self.N0_aii[a])
                magmom_v += (np.einsum('vij, ij -> v', M_vii,
                                       self.delta_aiiL[a][:, :, 0]) *
                             sqrt(4 * pi))
            domain_comm.sum(magmom_av)
            domain_comm.sum(magmom_v)
            magmom_v += self.nt_sR.integrate()[1:]

        return magmom_v, magmom_av

    def write(self, writer):
        D_asp = self.D_asii.to_cpu().to_lower_triangle().gather()
        nt_sR = self.nt_sR.to_xp(np).gather()
        if D_asp is None:
            return

        writer.write(
            density=nt_sR.data * Bohr**-3,
            atomic_density_matrices=D_asp.data)


def atomic_occupation_numbers(setup,
                              magmom_v: Vector,
                              ncomponents: int,
                              hund: bool = False,
                              charge: float = 0.0):
    M = np.linalg.norm(magmom_v)
    nspins = min(ncomponents, 2)
    f_si = setup.calculate_initial_occupation_numbers(
        M, hund, charge=charge, nspins=nspins)

    if ncomponents == 1:
        pass
    elif ncomponents == 2:
        if magmom_v[2] < 0:
            f_si = f_si[::-1].copy()
    else:
        f_i = f_si.sum(0)
        fm_i = f_si[0] - f_si[1]
        f_si = np.zeros((4, len(f_i)))
        f_si[0] = f_i
        if M > 0:
            f_si[1:] = np.asarray(magmom_v)[:, np.newaxis] / M * fm_i

    return f_si



import pytest
import numpy as np
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline


@pytest.mark.ci
def test_lfc_second_derivative():
    gd = GridDescriptor([20, 16, 16], [(4, 2, 0), (0, 4, 0), (0, 0, 4)])
    spos_ac = np.array([[0.252, 0.15, 0.35], [0.503, 0.5, 0.5]])
    s = Spline(l=0, rmax=2.0, f_g=np.array([1, 0.9, 0.1, 0.0]))
    spline_aj = [[s], [s]]
    c = LFC(gd, spline_aj)
    c.set_positions(spos_ac)
    c_ai = c.dict(zero=True)
    if 1 in c_ai:
        c_ai[1][0] = 2.0
    psi = gd.zeros()
    c.add(psi, c_ai)

    d_avv = dict([(a, np.zeros((3, 3))) for a in c.my_atom_indices])
    c.second_derivative(psi, d_avv)

    if 0 in d_avv:
        print(d_avv[0])

    eps = 0.000001
    d_aiv = c.dict(derivative=True)
    pos_av = np.dot(spos_ac, gd.cell_cv)
    for v in range(3):
        pos_av[0, v] += eps
        c.set_positions(np.dot(pos_av, gd.icell_cv.T))
        c.derivative(psi, d_aiv)
        if 0 in d_aiv:
            d0_v = d_aiv[0][0].copy()
        pos_av[0, v] -= 2 * eps
        c.set_positions(np.dot(pos_av, gd.icell_cv.T))
        c.derivative(psi, d_aiv)
        if 0 in d_aiv:
            d0_v -= d_aiv[0][0]
            d0_v /= -2 * eps
            print(d0_v)
            d_avv[0][v] -= d0_v
        pos_av[0, v] += eps
    if 0 in d_avv:
        assert np.abs(d_avv[0]).max() < 1e-10



import pytest
import numpy as np
import numpy.random as ra
from gpaw.setup import create_setup
from gpaw.xc import XC
from gpaw.test import equal, gen


@pytest.mark.libxc
def test_xc_lxc_xcatom(in_tmp_dir):
    if 1:
        setups = {}
        for functional in [
            'LDA_X', 'LDA_X+LDA_C_PW', 'LDA_X+LDA_C_VWN', 'LDA_X+LDA_C_PZ',
            'GGA_X_PBE+GGA_C_PBE', 'GGA_X_PBE_R+GGA_C_PBE',
            'GGA_X_B88+GGA_C_P86', 'GGA_X_B88+GGA_C_LYP',
            'GGA_X_FT97_A+GGA_C_LYP']:
            s = gen('N', xcname=functional)
            setups[functional] = s

    tolerance = 0.000005  # libxc must reproduce old gpaw energies
    # zero Kelvin: in Hartree

    reference = {  # version 0.9.1
        'LDA_X+LDA_C_PW': 2.28836113207,  # 'LDA'
        'GGA_X_PBE+GGA_C_PBE': 2.3366049993,  # 'PBE'
        'GGA_X_PBE_R+GGA_C_PBE': 2.34496288319}  # 'revPBE'

    reference_libxc = {  # svnversion 5252
        'LDA_X': 1.95030600807,
        'LDA_X+LDA_C_PW': 2.23194461135,
        'LDA_X+LDA_C_VWN': 2.23297429824,
        'LDA_X+LDA_C_PZ': 2.23146045547,
        'GGA_X_PBE+GGA_C_PBE': 2.28208665019,
        'GGA_X_PBE_R+GGA_C_PBE': 2.29201920843,
        'GGA_X_B88+GGA_C_P86': 2.30508027546,
        'GGA_X_B88+GGA_C_LYP': 2.28183010548,
        'GGA_X_FT97_A+GGA_C_LYP': 2.26846048873}

    libxc_set = [
        'LDA_X', 'LDA_X+LDA_C_PW', 'LDA_X+LDA_C_VWN', 'LDA_X+LDA_C_PZ',
        'GGA_X_PBE+GGA_C_PBE', 'GGA_X_PBE_R+GGA_C_PBE',
        'GGA_X_B88+GGA_C_P86', 'GGA_X_B88+GGA_C_LYP',
        'GGA_X_FT97_A+GGA_C_LYP']

    x = 0.000001
    for xcname in libxc_set:
        ra.seed(8)
        xc = XC(xcname)
        s = create_setup('N', xc, setupdata=setups[xcname])
        ni = s.ni
        nii = ni * (ni + 1) // 2
        D_p = 0.1 * ra.random(nii) + 0.4
        H_p = np.zeros(nii)

        E1 = xc.calculate_paw_correction(s,
                                         D_p.reshape(1, -1),
                                         H_p.reshape(1, -1))
        dD_p = x * ra.random(nii)
        D_p += dD_p
        dE = np.dot(H_p, dD_p) / x
        E2 = xc.calculate_paw_correction(s, D_p.reshape(1, -1))
        print(xcname, dE, (E2 - E1) / x)
        equal(dE, (E2 - E1) / x, 0.003)

        E2s = xc.calculate_paw_correction(
            s,
            np.array([0.5 * D_p, 0.5 * D_p]),
            np.array([H_p, H_p]))
        print(E2, E2s)
        equal(E2, E2s, 1.0e-12)

        if xcname in reference:  # compare with old gpaw
            print('A:', E2, reference[xcname])
            equal(E2, reference[xcname], tolerance)

        if xc in reference_libxc:  # compare with reference libxc
            print('B:', E2, reference_libxc[xcname])
            equal(E2, reference_libxc[xcname], tolerance)

        D_sp = 0.1 * ra.random((2, nii)) + 0.2
        H_sp = np.zeros((2, nii))

        E1 = xc.calculate_paw_correction(s, D_sp, H_sp)
        dD_sp = x * ra.random((2, nii))
        D_sp += dD_sp
        dE = np.dot(H_sp.ravel(), dD_sp.ravel()) / x
        E2 = xc.calculate_paw_correction(s, D_sp, H_sp)
        print(dE, (E2 - E1) / x)
        equal(dE, (E2 - E1) / x, 0.005)



import gpaw.solvation as solv
from ase import Atoms
import numpy as np
from gpaw.solvation.cavity import BAD_RADIUS_MESSAGE


def test_solvation_nan_radius():
    atoms = Atoms('H')
    atoms.center(vacuum=3.0)
    kwargs = solv.get_HW14_water_kwargs()

    def rfun(a):
        return [np.nan]

    kwargs['cavity'].effective_potential.atomic_radii = rfun
    atoms.calc = solv.SolvationGPAW(xc='LDA', h=0.24, **kwargs)
    try:
        atoms.get_potential_energy()
    except ValueError as error:
        if not error.args[0] == BAD_RADIUS_MESSAGE:
            raise
    else:
        raise AssertionError("Expected ValueError")



import numpy as np


def fftshift(a_Q, axes):
    from gpaw.gpu.cpupy import ndarray
    return ndarray(np.fft.fftshift(a_Q._data, axes=axes))


def ifftshift(a_Q, axes):
    from gpaw.gpu.cpupy import ndarray
    return ndarray(np.fft.ifftshift(a_Q._data, axes=axes))



# flake8: noqa
import os

import io

import numpy as np

import ase.units

import gpaw.mpi

from gpaw.utilities.tools import coordinates
from gpaw.utilities.tools import pick
from gpaw.utilities import pack

from gpaw.fd_operators import Gradient


class KohnShamSingleExcitation:
    def __init__(self, gs_calc, kpt_ind, occ_ind, unocc_ind):
        self.calc = gs_calc
        self.kpt_ind = kpt_ind
        self.spin_ind = self.calc.wfs.kpt_u[self.kpt_ind].s
        self.occ_ind = occ_ind
        self.unocc_ind = unocc_ind
        self.energy_diff = None
        self.pop_diff = None
        self.dip_mom_r = None
        self.magn_mom = None
        self.pair_density = None

    # Calculates pair density of (i,p) transition (and given kpt)
    # dnt_Gp: pair density without compensation charges on coarse grid
    # dnt_gp: pair density without compensation charges on fine grid      (XC)
    # drhot_gp: pair density with compensation charges on fine grid  (poisson)
    def calculate_pair_density(self, dnt_Gip, dnt_gip, drhot_gip):
        if self.pair_density is None:
            # FIXME: USE EXISTING PairDensity method
            self.pair_density = LRiPairDensity(self.calc.density)
        self.pair_density.initialize(self.calc.wfs.kpt_u[self.kpt_ind],
                                     self.occ_ind, self.unocc_ind)
        self.pair_density.get(dnt_Gip, dnt_gip, drhot_gip)

    def __str__(self):
        if self.dip_mom_r is not None and self.dip_mom_v is not None and self.magn_mom is not None:
            str = "# KS single excitation from state %05d to state %05d: dE_pi = %18.12lf, f_pi = %18.12lf,  dmr_ip = (%18.12lf, %18.12lf, %18.12lf), dmv_ip = (%18.12lf, %18.12lf, %18.12lf), dmm_ip = %18.12lf" \
                % ( self.occ_ind, \
                        self.unocc_ind, \
                        self.energy_diff, \
                        self.pop_diff, \
                        self.dip_mom_r[0], self.dip_mom_r[1], self.dip_mom_r[2], \
                        self.dip_mom_v[0], self.dip_mom_v[1], self.dip_mom_v[2], \
                        self.magn_mom )
        elif self.energy_diff is not None and self.pop_diff is not None:
            str = "# KS single excitation from state %05d to state %05d: dE_pi = %18.12lf, f_pi = %18.12lf" \
                % ( self.occ_ind, \
                        self.unocc_ind,  \
                        self.energy_diff, \
                        self.pop_diff )
        elif self.occ_ind is not None and self.unocc_ind is not None:
            str = "# KS single excitation from state %05d to state %05d" \
                % ( self.occ_ind, self.unocc_ind )
        else:
            raise RuntimeError("Uninitialized KSSingle")
        return str


class KohnShamSingles:
    def __init__(self, basefilename, gs_calc, kpt_index, min_occ, max_occ,
                 min_unocc, max_unocc, max_energy_diff, min_pop_diff, lr_comms,
                 txt):
        self.basefilename = basefilename
        self.calc = gs_calc
        self.min_occ = min_occ
        self.max_occ = max_occ
        self.min_unocc = min_unocc
        self.max_unocc = max_unocc
        self.max_energy_diff = max_energy_diff
        self.min_pop_diff = min_pop_diff
        self.kpt_ind = kpt_index
        self.lr_comms = lr_comms
        self.txt = txt

        self.kss_list = None
        self.kss_list_ready = False

        self.kss_prop = None
        self.kss_prop_ready = False

    def update_list(self):
        # shorthands
        eps_n = self.calc.wfs.kpt_u[self.kpt_ind].eps_n  # eigen energies
        f_n = self.calc.wfs.kpt_u[self.kpt_ind].f_n  # occupations

        # Create Kohn-Sham single excitation list with energy filter
        old_kss_list = self.kss_list  # save old list for later
        self.kss_list = []  # create a completely new list
        # Occupied loop
        for i in range(self.min_occ, self.max_occ + 1):
            # Unoccupied loop
            for p in range(self.min_unocc, self.max_unocc + 1):
                deps_pi = eps_n[p] - eps_n[i]  # energy diff
                df_ip = f_n[i] - f_n[p]  # population diff
                # Filter it
                if (np.abs(deps_pi) <= self.max_energy_diff
                        and df_ip > self.min_pop_diff):
                    # i, p, deps, df, mur, muv, magn
                    kss = KohnShamSingleExcitation(self.calc, self.kpt_ind, i,
                                                   p)
                    kss.energy_diff = deps_pi
                    kss.pop_diff = df_ip
                    self.kss_list.append(kss)

        # Sort by energy diff:
        self.kss_list = sorted(self.kss_list, key=lambda x: x.energy_diff)

        # Remove old transitions and add new
        # BUT only add to the end of the list (otherwise lower triangle
        # matrix is not filled completely)
        if old_kss_list is not None:
            new_kss_list = self.kss_list  # required list
            self.kss_list = []  # final list with correct order
            # Old list first
            # If in new and old lists
            for kss_o in old_kss_list:
                found = False
                for kss_n in new_kss_list:
                    if (kss_o.occ_ind == kss_n.occ_ind
                            and kss_o.unocc_ind == kss_n.unocc_ind):
                        found = True
                        break
                if found:
                    self.kss_list.append(kss_o)  # Found, add to final list
                else:
                    pass  # else drop

            # Now old transitions which are not in new list where dropped

            # If only in new list
            app_list = []
            for kss_n in new_kss_list:
                found = False
                for kss in self.kss_list:
                    if (kss.occ_ind == kss_n.occ_ind
                            and kss.unocc_ind == kss_n.unocc_ind):
                        found = True
                        break
                if not found:
                    app_list.append(kss_n)  # Not found, add to final list
                else:
                    pass  # else skip to avoid duplicates

            # Create the final list
            self.kss_list += app_list

        self.txt.write('Number of electron-hole pairs: %d\n' %
                       len(self.kss_list))
        self.txt.write('Maximum energy difference:     %6.3lf\n' %
                       (self.kss_list[-1].energy_diff * ase.units.Hartree))

        # Prevent repeated work
        self.kss_list_ready = True

    def read(self):
        # Read KS_singles file if exists
        # occ_index | unocc index | energy diff | population diff |
        #   dmx | dmy | dmz | magnx | magny | magnz
        kss_file = self.basefilename + '.KS_singles'
        if os.path.exists(kss_file) and os.path.isfile(kss_file):
            self.kss_list = []

            # root reads and broadcasts
            # a bit ugly and slow but works
            data = None
            if self.lr_comms.parent_comm.rank == 0:
                with open(kss_file, 'r', 1024 * 1024) as fd:
                    data = fd.read()
            data = gpaw.mpi.broadcast_string(data,
                                             root=0,
                                             comm=self.lr_comms.parent_comm)

            # just parse
            for line in io.StringIO(data):
                line = line.split()
                i, p = int(line[0]), int(line[1])
                ediff, fdiff = float(line[2]), float(line[3])
                dm = np.array([float(line[4]), float(line[5]), float(line[6])])
                mm = np.array([float(line[7]), float(line[8]), float(line[9])])
                kss = KohnShamSingleExcitation(self.calc, self.kpt_ind, i, p)
                kss.energy_diff = ediff
                kss.pop_diff = fdiff
                kss.dip_mom_r = dm
                kss.magn_mom = mm
                #assert self.index_of_kss(i,p) is None, 'KS transition %d->%d found twice in KS_singles files.' % (i,p)
                self.kss_list.append(kss)

            # if none read
            if len(self.kss_list) <= 0:
                self.kss_list = None

    # Calculate dipole moment and magnetic moment for noninteracting
    # Kohn-Sham transitions
    # FIXME: Do it in parallel, currently doing repeated work
    #        Should distribute kss_list over eh_comm
    def calculate(self):
        # Check that kss_list is up-to-date
        if not self.kss_list_ready:
            self.update_list()
            self.kss_prop_ready = False

        # Check if we already have properties for all singles
        if self.kss_prop_ready:
            return
        self.kss_prop_ready = True
        for kss_ip in self.kss_list:
            if kss_ip.dip_mom_r is None or kss_ip.magn_mom is None:
                self.kss_prop_ready = False
                break
        # If had, done
        if self.kss_prop_ready:
            return

        #self.timer.start('Calculate KS properties')

        # Init pair densities
        dnt_Gip = self.calc.wfs.gd.empty()
        dnt_gip = self.calc.density.finegd.empty()
        drhot_gip = self.calc.density.finegd.empty()

        # Init gradients of wfs
        grad_psit2_G = [
            self.calc.wfs.gd.empty(),
            self.calc.wfs.gd.empty(),
            self.calc.wfs.gd.empty()
        ]

        # Init gradient operators
        grad = []
        dtype = pick(self.calc.wfs.kpt_u[self.kpt_ind].psit_nG, 0).dtype
        for c in range(3):
            grad.append(Gradient(self.calc.wfs.gd, c, dtype=dtype, n=2))

        # Coordinate vector r
        R0 = 0.5 * np.diag(self.calc.wfs.gd.cell_cv)
        # R0 = np.array([0.0,0.0,0.0]) # old_version
        #print 'R0', R0
        r_cG, r2_G = coordinates(self.calc.wfs.gd, origin=R0)
        r_cg, r2_g = coordinates(self.calc.density.finegd, origin=R0)

        # Loop over all KS single excitations
        #
        # Transition dipole moment, mu_ip = <p| (-e r) |i>
        # Magnetic transition dipole, m_ip = -(1/2c) <i|L|p>
        # For total m_0I = -m_I0 = -(m_0I)^*, but not for m_ip (right?)
        # R_0I = Im[mu_0I * m_0I]
        # mu_ip^0I =  omega_0I^(-1/2) D_ip      S^(-1/2) F_0I
        # m_ip^0I  = -omega_0I^(+1/2) M_ip C^-1 S^(+1/2) F_0I
        #
        # S_ip,ip = - (eps_p - eps_i) / (n_p - n_i)    (note: n_p < n_i)
        # C_ip,ip = 1 / (n_p - n_i)
        #
        # See:
        # WIREs Comput Mol Sci 2012, 2: 150-166 doi: 10.1002/wcms.55
        # J. Chem. Phys., Vol. 116, 6930 (2002)
        for kss_ip in self.kss_list:
            # If have dipole moment and magnetic moment, already done and skip
            if (kss_ip.dip_mom_r is not None and kss_ip.magn_mom is not None):
                continue

            # Transition dipole moment, mu_ip = <p| (-r) |i>
            kss_ip.calculate_pair_density(dnt_Gip, dnt_gip, drhot_gip)
            #kss_ip.dip_mom_r = self.calc.density.finegd.calculate_dipole_moment(drhot_gip)
            #kss_ip.dip_mom_r = self.calc.density.finegd.calculate_dipole_moment(drhot_gip)
            kss_ip.dip_mom_r = np.zeros(3)
            kss_ip.dip_mom_r[0] = -self.calc.density.finegd.integrate(
                r_cg[0] * drhot_gip)
            kss_ip.dip_mom_r[1] = -self.calc.density.finegd.integrate(
                r_cg[1] * drhot_gip)
            kss_ip.dip_mom_r[2] = -self.calc.density.finegd.integrate(
                r_cg[2] * drhot_gip)

            # Magnetic transition dipole, m_ip = -(1/2c) <i|L|p> = i/2c <i|r x p|p>
            # see Autschbach et al., J. Chem. Phys., 116, 6930 (2002)

            # Gradients
            for c in range(3):
                grad[c].apply(kss_ip.pair_density.psit2_G, grad_psit2_G[c],
                              self.calc.wfs.kpt_u[self.kpt_ind].phase_cd)

            # <psi1|r x grad|psi2>
            #    i  j  k
            #    x  y  z   = (y pz - z py)i + (z px - x pz)j + (x py - y px)
            #    px py pz
            rxnabla_g = np.zeros(3)
            rxnabla_g[0] = self.calc.wfs.gd.integrate(
                kss_ip.pair_density.psit1_G *
                (r_cG[1] * grad_psit2_G[2] - r_cG[2] * grad_psit2_G[1]))
            rxnabla_g[1] = self.calc.wfs.gd.integrate(
                kss_ip.pair_density.psit1_G *
                (r_cG[2] * grad_psit2_G[0] - r_cG[0] * grad_psit2_G[2]))
            rxnabla_g[2] = self.calc.wfs.gd.integrate(
                kss_ip.pair_density.psit1_G *
                (r_cG[0] * grad_psit2_G[1] - r_cG[1] * grad_psit2_G[0]))

            # augmentation contributions to magnetic moment
            # <psi1| r x nabla |psi2> = <psi1| (r-Ra+Ra) x nabla |psi2>
            #                         = <psi1| (r-Ra) x nabla |psi2> + Ra x <psi1| nabla |psi2>
            rxnabla_a = np.zeros(3)
            # <psi1| (r-Ra) x nabla |psi2>
            for a, P_ni in self.calc.wfs.kpt_u[self.kpt_ind].P_ani.items():
                Pi_i = P_ni[kss_ip.occ_ind]
                Pp_i = P_ni[kss_ip.unocc_ind]
                rxnabla_iiv = self.calc.wfs.setups[a].rxnabla_iiv
                for c in range(3):
                    for i1, Pi in enumerate(Pi_i):
                        for i2, Pp in enumerate(Pp_i):
                            rxnabla_a[c] += Pi * Pp * rxnabla_iiv[i1, i2, c]

            self.calc.wfs.gd.comm.sum(rxnabla_a)  # sum up from different procs

            # Ra x <psi1| nabla |psi2>
            Rxnabla_a = np.zeros(3)
            for a, P_ni in self.calc.wfs.kpt_u[self.kpt_ind].P_ani.items():
                Pi_i = P_ni[kss_ip.occ_ind]
                Pp_i = P_ni[kss_ip.unocc_ind]
                nabla_iiv = self.calc.wfs.setups[a].nabla_iiv
                Ra = (self.calc.atoms[a].position / ase.units.Bohr) - R0
                for i1, Pi in enumerate(Pi_i):
                    for i2, Pp in enumerate(Pp_i):
                        # (y pz - z py)i + (z px - x pz)j + (x py - y px)k
                        Rxnabla_a[0] += Pi * Pp * (
                            Ra[1] * nabla_iiv[i1, i2, 2] -
                            Ra[2] * nabla_iiv[i1, i2, 1])
                        Rxnabla_a[1] += Pi * Pp * (
                            Ra[2] * nabla_iiv[i1, i2, 0] -
                            Ra[0] * nabla_iiv[i1, i2, 2])
                        Rxnabla_a[2] += Pi * Pp * (
                            Ra[0] * nabla_iiv[i1, i2, 1] -
                            Ra[1] * nabla_iiv[i1, i2, 0])

            self.calc.wfs.gd.comm.sum(Rxnabla_a)  # sum up from different procs

            #print (kss_ip.occ_ind, kss_ip.unocc_ind), kss_ip.dip_mom_r, rxnabla_g, rxnabla_a, Rxnabla_a

            # m_ip = -1/2c <i|r x p|p> = i/2c <i|r x nabla|p>
            # just imaginary part!!!
            kss_ip.magn_mom = ase.units.alpha / 2. * (rxnabla_g + rxnabla_a +
                                                      Rxnabla_a)

        # Wait... to avoid io problems, and write KS_singles file
        self.lr_comms.parent_comm.barrier()
        if self.lr_comms.parent_comm.rank == 0:
            self.kss_file = open(self.basefilename + '.KS_singles', 'w')
            for kss_ip in self.kss_list:
                format = '%08d %08d  %18.12lf %18.12lf  '
                format += '%18.12lf %18.12lf %18.12lf '
                format += '%18.12lf %18.12lf %18.12lf\n'
                self.kss_file.write(
                    format %
                    (kss_ip.occ_ind, kss_ip.unocc_ind, kss_ip.energy_diff,
                     kss_ip.pop_diff, kss_ip.dip_mom_r[0], kss_ip.dip_mom_r[1],
                     kss_ip.dip_mom_r[2], kss_ip.magn_mom[0],
                     kss_ip.magn_mom[1], kss_ip.magn_mom[2]))
            self.kss_file.close()
        self.lr_comms.parent_comm.barrier()

        self.kss_prop_ready = True  # avoid repeated work
        #self.timer.stop('Calculate KS properties')


# Small utility class

# FIXME: USE EXISTING METHOD
class LRiPairDensity:
    """Pair density calculator class."""
    def __init__(self, density):
        """Initialization needs density instance"""
        self.density = density

    def initialize(self, kpt, n1, n2):
        """Set wave function indices."""
        self.n1 = n1
        self.n2 = n2
        self.spin = kpt.s
        self.P_ani = kpt.P_ani
        self.psit1_G = pick(kpt.psit_nG, n1)
        self.psit2_G = pick(kpt.psit_nG, n2)

    def get(self, nt_G, nt_g, rhot_g):
        """Get pair densities.

        nt_G
          Pair density without compensation charges on the coarse grid

        nt_g
          Pair density without compensation charges on the fine grid

        rhot_g
          Pair density with compensation charges on the fine grid
        """
        # Coarse grid product
        np.multiply(self.psit1_G.conj(), self.psit2_G, nt_G)
        # Interpolate to fine grid
        self.density.interpolator.apply(nt_G, nt_g)

        # Determine the compensation charges for each nucleus
        Q_aL = {}
        for a, P_ni in self.P_ani.items():
            assert P_ni.dtype == float
            # Generate density matrix
            P1_i = P_ni[self.n1]
            P2_i = P_ni[self.n2]
            D_ii = np.outer(P1_i.conj(), P2_i)
            # allowed to pack as used in the scalar product with
            # the symmetric array Delta_pL
            D_p = pack(D_ii)
            #FIXME: CHECK THIS D_p  = pack(D_ii, tolerance=1e30)

            # Determine compensation charge coefficients:
            Q_aL[a] = np.dot(D_p, self.density.setups[a].Delta_pL)

        # Add compensation charges
        rhot_g[:] = nt_g[:]
        self.density.ghat.add(rhot_g, Q_aL)
        #print 'dens', self.density.finegd.integrate(rhot_g)



from gpaw import GPAW
from gpaw.eigensolvers.rmmdiis import RMMDIIS
from ase import Atoms
from gpaw.test import equal


def test_eigen_keep_htpsit():
    atoms = Atoms('H')
    atoms.center(3.0)

    convergence = {'eigenstates': 1e-2, 'density': 1e-2}
    # Keep htpsit
    calc = GPAW(nbands=2,
                eigensolver=RMMDIIS(keep_htpsit=True),
                convergence=convergence,
                maxiter=20)
    atoms.calc = calc
    e0 = atoms.get_potential_energy()

    # Do not keep htpsit
    calc = GPAW(nbands=2,
                eigensolver=RMMDIIS(keep_htpsit=False),
                convergence=convergence,
                maxiter=20)
    atoms.calc = calc
    e1 = atoms.get_potential_energy()

    equal(e0, e1, 1e-12)



import numpy as np
from gpaw.utilities.blas import mmm


class HilbertTransform:
    def __init__(self, omega_w, eta, timeordered=False, gw=False,
                 blocksize=500):
        """Analytic Hilbert transformation using linear interpolation.

        Hilbert transform::

           oo
          /           1                1
          |dw' (-------------- - --------------) S(w').
          /     w - w' + i eta   w + w' + i eta
          0

        With timeordered=True, you get::

           oo
          /           1                1
          |dw' (-------------- - --------------) S(w').
          /     w - w' - i eta   w + w' + i eta
          0

        With gw=True, you get::

           oo
          /           1                1
          |dw' (-------------- + --------------) S(w').
          /     w - w' + i eta   w + w' + i eta
          0

        """

        self.blocksize = blocksize

        if timeordered:
            self.H_ww = self.H(omega_w, -eta) + self.H(omega_w, -eta, -1)
        elif gw:
            self.H_ww = self.H(omega_w, eta) - self.H(omega_w, -eta, -1)
        else:
            self.H_ww = self.H(omega_w, eta) + self.H(omega_w, -eta, -1)

    def H(self, o_w, eta, sign=1):
        """Calculate transformation matrix.

        With s=sign (+1 or -1)::

                        oo
                       /       dw'
          X (w, eta) = | ---------------- S(w').
           s           / s w - w' + i eta
                       0

        Returns H_ij so that X_i = np.dot(H_ij, S_j), where::

            X_i = X (omega_w[i]) and S_j = S(omega_w[j])
                   s
        """

        nw = len(o_w)
        H_ij = np.zeros((nw, nw), complex)
        do_j = o_w[1:] - o_w[:-1]
        for i, o in enumerate(o_w):
            d_j = o_w - o * sign
            y_j = 1j * np.arctan(d_j / eta) + 0.5 * np.log(d_j**2 + eta**2)
            y_j = (y_j[1:] - y_j[:-1]) / do_j
            H_ij[i, :-1] = 1 - (d_j[1:] - 1j * eta) * y_j
            H_ij[i, 1:] -= 1 - (d_j[:-1] - 1j * eta) * y_j
        return H_ij

    def __call__(self, S_wx):
        """Inplace transform"""
        B_wx = S_wx.reshape((len(S_wx), -1))
        nw, nx = B_wx.shape
        tmp_wx = np.zeros((nw, min(nx, self.blocksize)), complex)
        for x in range(0, nx, self.blocksize):
            b_wx = B_wx[:, x:x + self.blocksize]
            c_wx = tmp_wx[:, :b_wx.shape[1]]
            mmm(1.0, self.H_ww, 'N', b_wx, 'N', 0.0, c_wx)
            b_wx[:] = c_wx


class GWHilbertTransforms:
    """Helper class which wraps two transforms using contiguous array.

    (This slightly speeds up things.)"""
    def __init__(self, omega_w, eta):
        self.htp = htp = HilbertTransform(omega_w, eta, gw=True)
        self.htm = htm = HilbertTransform(omega_w, -eta, gw=True)
        self._stacked_H_nww = np.array([htp.H_ww, htm.H_ww])

    def __call__(self, A_wGG):
        # (Note: This effectively duplicates the Hilbert call)
        nw = len(A_wGG)
        H_xw = self._stacked_H_nww.reshape(-1, nw)
        A_wy = A_wGG.reshape(nw, -1)
        tmp_xy = np.zeros((H_xw.shape[0], A_wy.shape[1]), complex)
        # gemm(1.0, A_wy, H_xw, 0.0, tmp_xy)
        mmm(1.0, H_xw, 'N', A_wy, 'N', 0.0, tmp_xy)
        return tmp_xy.reshape((2, *A_wGG.shape))



import numpy as np

from gpaw.utilities.blas import mmm
from gpaw.utilities import unpack


class BaseAtomicCorrection:
    name = 'base'
    description = 'base class for atomic corrections with LCAO'

    def __init__(self, dS_aii, Mstart, Mstop):
        self.dS_aii = dS_aii
        self.Mstart = Mstart
        self.Mstop = Mstop

    @staticmethod
    def get_dS(atom_partition, setups):
        dO_aii = atom_partition.arraydict([setup.dO_ii.shape
                                           for setup in setups])
        for a in dO_aii:
            dO_aii[a][:] = setups[a].dO_ii
        return dO_aii

    def calculate_hamiltonian(self, kpt, dH_asp, H_MM, yy):
        dH_aii = dH_asp.partition.arraydict(self.dS_aii.shapes_a,
                                            dtype=dH_asp.dtype)

        for a in dH_asp:
            dH_aii[a] = yy * unpack(dH_asp[a][kpt.s])

        self.calculate(kpt.q, dH_aii, H_MM, self.Mstart, self.Mstop)

    def add_overlap_correction(self, S_qMM):
        for q, S_MM in enumerate(S_qMM):
            self.calculate(q, self.dS_aii, S_MM, self.Mstart, self.Mstop)

    def calculate(self, q, dX_aii, X_MM):
        raise NotImplementedError


class DenseAtomicCorrection(BaseAtomicCorrection):
    name = 'dense'
    description = 'dense with blas'

    def __init__(self, P_aqMi, dS_aii, Mstart, Mstop):
        BaseAtomicCorrection.__init__(self, dS_aii, Mstart, Mstop)
        self.P_aqMi = P_aqMi

    @classmethod
    def new_from_wfs(cls, wfs):
        return cls(wfs.P_aqMi, cls.get_dS(wfs.atom_partition, wfs.setups),
                   wfs.ksl.Mstart, wfs.ksl.Mstop)

    def calculate(self, q, dX_aii, X_MM, Mstart, Mstop):
        dtype = X_MM.dtype
        P_aqMi = self.P_aqMi

        # P_aqMi is distributed over domains (a) and bands (M).
        # Hence the correction X_MM = sum(P dX P) includes contributions
        # only from local atoms; the result must be summed over gd.comm
        # to get all 'a' contributions, and it will be locally calculated
        # only on the local slice of bands.
        for a, dX_ii in dX_aii.items():
            P_Mi = P_aqMi[a][q]
            assert dtype == P_Mi.dtype
            dXP_iM = np.zeros((dX_ii.shape[1], P_Mi.shape[0]), dtype)
            # (ATLAS can't handle uninitialized output array)
            mmm(1.0, np.asarray(dX_ii, dtype), 'N', P_Mi, 'C', 0.0, dXP_iM)
            mmm(1.0, P_Mi[Mstart:Mstop], 'N', dXP_iM, 'N', 1.0, X_MM)

    def calculate_projections(self, wfs, kpt):
        for a, P_ni in kpt.P_ani.items():
            # ATLAS can't handle uninitialized output array:
            P_ni.fill(117)
            mmm(1.0, kpt.C_nM, 'N', wfs.P_aqMi[a][kpt.q], 'N', 0.0, P_ni)


class SparseAtomicCorrection(BaseAtomicCorrection):
    name = 'sparse'
    description = 'sparse using scipy'

    def __init__(self, Psparse_qIM, P_indices, dS_aii, Mstart, Mstop,
                 tolerance=1e-12):
        BaseAtomicCorrection.__init__(self, dS_aii, Mstart, Mstop)
        self.Psparse_qIM = Psparse_qIM
        self.P_indices = P_indices
        # We currently don't use tolerance although we could speed things
        # up that way.
        #
        # Tolerance is for zeroing elements very close to zero, which
        # often increases sparsity somewhat, even for very small values.
        self.tolerance = tolerance

    @classmethod
    def new_from_wfs(cls, wfs):
        return cls(wfs.P_qIM, wfs.setups.projector_indices(),
                   cls.get_dS(wfs.atom_partition, wfs.setups),
                   wfs.ksl.Mstart, wfs.ksl.Mstop)

    def calculate(self, q, dX_aii, X_MM, Mstart, Mstop):
        P_indices = self.P_indices
        nI = P_indices.max

        import scipy.sparse as sparse
        dXsparse_II = sparse.dok_matrix((nI, nI), dtype=X_MM.dtype)
        for a in dX_aii:
            I1, I2 = P_indices[a]
            dXsparse_II[I1:I2, I1:I2] = dX_aii[a]
        dXsparse_II = dXsparse_II.tocsr()

        Psparse_IM = self.Psparse_qIM[q]
        Psparse_MI = Psparse_IM[:, Mstart:Mstop].transpose().conj()
        Xsparse_MM = Psparse_MI.dot(dXsparse_II.dot(Psparse_IM))
        X_MM[:, :] += Xsparse_MM.todense()

    def calculate_projections(self, wfs, kpt):
        P_indices = self.P_indices
        P_In = self.Psparse_qIM[kpt.q].dot(kpt.C_nM.T.conj())
        for a in kpt.P_ani:
            I1, I2 = P_indices[a]
            kpt.P_ani[a][:, :] = P_In[I1:I2, :].T.conj()



from pathlib import Path
import pytest
from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft.dielectric import get_dielectric, dielectric


@pytest.fixture
def H2():
    R = 0.7  # approx. experimental bond length
    a = 3.0
    c = 4.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    H2.calc = GPAW(h=0.25, nbands=3, spinpol=False, txt=None)
    H2.get_potential_energy()
    return H2


@pytest.fixture
def lrtddft(H2):
    exlst = LrTDDFT(txt=None)
    exlst.calculate(H2)
    return exlst


@pytest.mark.lrtddft
def test_get(H2, lrtddft):
    energies, eps1, eps2, N, K, R = get_dielectric(lrtddft, H2.get_volume())
    for res in [eps1, eps2, N, K, R]:
        assert energies.shape == res.shape
    assert (energies >= 0).all()


@pytest.mark.lrtddft
def test_write(H2, lrtddft, in_tmp_dir):
    fname = 'dielectric.dat'
    dielectric(lrtddft, H2.get_volume(), fname)
    assert Path(fname).is_file()



from gpaw.solvation.cavity import get_pbc_positions
from ase.build import molecule
from ase.units import Bohr
import numpy as np


def test_solvation_pbc_pos_repeat():
    def get_nrepeats(nrepeats_c):
        return np.prod(np.array(nrepeats_c) * 2 + 1)

    def check_valid_repeat(pos_v, rpos_v, nrepeats_c, cell_cv):
        nrepeats_c = np.array(nrepeats_c)
        diff_v = rpos_v - pos_v
        actual_nrepeats_c = np.dot(diff_v, np.linalg.inv(cell_cv))
        r_actual_nrepeats_c = np.around(actual_nrepeats_c)
        # check for integer multiple
        assert np.allclose(r_actual_nrepeats_c, actual_nrepeats_c)
        actual_nrepeats_c = np.abs(r_actual_nrepeats_c.astype(int))
        # check for too large repeats
        assert (actual_nrepeats_c <= nrepeats_c).all()

    def check(pos_av, rpos_aav, nrepeats_c, cell_cv):
        assert len(pos_av) == len(rpos_aav)
        total_repeats = get_nrepeats(nrepeats_c)
        for a, pos_v in enumerate(pos_av):
            rpos_av = rpos_aav[a]
            assert len(rpos_av) == total_repeats
            for rpos_v in rpos_av:
                check_valid_repeat(pos_v, rpos_v, nrepeats_c, cell_cv)

    cells = (
        ((10., .0, .0), (.0, 12., .0), (.0, .0, 12.)),  # orthogonal
        ((10., 1., .0), (.0, 12., 1.), (1., .0, 11.5)))  # non-orthogonal

    for cell in cells:
        atoms = molecule('H2O')
        atoms.center(vacuum=5.)
        atoms.set_cell(cell)
        pos_av = atoms.positions / Bohr
        cell_cv = atoms.get_cell() / Bohr
        Rcell_c = np.sqrt(np.sum(cell_cv ** 2, axis=1))

        # non periodic should never repeat
        atoms.pbc = np.zeros((3, ), dtype=bool)
        pos_aav = get_pbc_positions(atoms, 1e10 / Bohr)
        check(pos_av, pos_aav, (0, 0, 0), cell_cv)

        # periodic, zero cutoff should not repeat
        atoms.pbc = np.ones((3, ), dtype=bool)
        pos_aav = get_pbc_positions(atoms, .0 / Bohr)
        check(pos_av, pos_aav, (0, 0, 0), cell_cv)

        # periodic, cutoff <= cell size should repeat once
        atoms.pbc = np.ones((3, ), dtype=bool)
        pos_aav = get_pbc_positions(atoms, 0.99 * Rcell_c.min())
        check(pos_av, pos_aav, (1, 1, 1), cell_cv)

        # periodic, cutoff > cell size should repeat twice
        atoms.pbc = np.ones((3, ), dtype=bool)
        pos_aav = get_pbc_positions(atoms, 1.01 * Rcell_c.max())
        check(pos_av, pos_aav, (2, 2, 2), cell_cv)

        # mixed bc, cutoff > 2 * cell size should repeat three times
        for i in range(8):
            pbc = np.array([int(p) for p in np.binary_repr(i, 3)])
            atoms.pbc = pbc
            pos_aav = get_pbc_positions(atoms, 2.01 * Rcell_c.max())
            check(pos_av, pos_aav, pbc * 3, cell_cv)
